"""This module contains the base class for plugins which loads
devices credentials
"""
import logging
from abc import abstractmethod
from typing import Dict, List

from suzieq.poller.controller.base_controller_plugin import \
    InventoryPluginModel, ControllerPlugin
from suzieq.shared.exceptions import InventorySourceError

logger = logging.getLogger(__name__)


def check_credentials(inventory: Dict):
    """Checks if all the credentials are set
    """

    # check if all devices has credentials
    no_cred_nodes = [
        f"{d.get('namespace')}.{d.get('address')}"
        for d in inventory.values()
        if not d.get('username', None) or
        not (d.get('password') or d.get('ssh_keyfile'))
    ]
    if no_cred_nodes:
        raise InventorySourceError(
            'No credentials to log into the following nodes: '
            f'{no_cred_nodes}'
        )


class CredentialLoaderModel(InventoryPluginModel):
    """Model for credential loader validation
    """


class CredentialLoader(ControllerPlugin):
    """Base class used to import device credentials from different
    sources
    """

    def __init__(self, init_data: Dict, validate: bool = True) -> None:
        super().__init__(init_data, validate)

        self._cred_format = [
            'username',
            'password',
            'ssh_keyfile',
            'passphrase',
            'enable_password'
        ]

        self._optional_cred_keys = [
            'enable_password',
        ]

        self._data: CredentialLoaderModel = None
        self.init(init_data)

    @property
    def name(self) -> str:
        """Name of the source set in the inventory file

        Returns:
            str: name of the source
        """
        return self._data.name

    @classmethod
    def default_type(cls) -> str:
        return 'static'

    @classmethod
    def get_data_model(cls):
        return CredentialLoaderModel

    def init(self, init_data: Dict):
        """Initialize the object

        Args:
            init_data (Dict): data used to initialize the object
        """
        if self._validate:
            self._data = self.get_data_model()(**init_data)
        else:
            self._data = self.get_data_model().construct(**init_data)
        if not self._data:
            raise InventorySourceError(
                'input_data was not loaded correctly')

    @abstractmethod
    def load(self, inventory: Dict[str, Dict]):
        """Loads the credentials inside the inventory

        Args:
            inventory (Dict[str, Dict]): inventory to update
        """

    def write_credentials(self, device: Dict, credentials: Dict[str, Dict]):
        """write and validate input credentials for a device

        Args:
            device (Dict): device to add credentials
            credentials (Dict[str, Dict]): device's credentials

        Raises:
            InventorySourceError: Invalid credentials
        """
        missing_keys = self._validate_credentials(credentials)
        if missing_keys:
            raise InventorySourceError(
                f'Invalid credentials: missing keys {missing_keys}')
        device.update(credentials)

    def _validate_credentials(self, credentials: Dict) -> List[str]:
        """Checks if provided credentials are valid

        Args:
            credentials (Dict): device credentials

        Raises:
            RuntimeError: Unexpected key

        Returns:
            List[str]: list of missing fields
        """
        cred_keys = set(self._cred_format)
        for key, value in credentials.items():
            if key in cred_keys:
                # 'passphrase' is valid also with None value
                # Also 'password' and 'ssh_keyfile' with None value are valid
                # but only if at least one of them has a not None value
                if value or key == 'passphrase':
                    cred_keys.remove(key)
            else:
                raise InventorySourceError(
                    f'Unexpected key {key} in credentials')

        # One between password or ssh_keyfile must be defines
        if 'password' in cred_keys and 'ssh_keyfile' in cred_keys:
            cred_keys.remove('password')
            cred_keys.remove('ssh_keyfile')
            ret = list(cred_keys)
            ret.append('password or ssh_keyfile')
            return ret

        # if we arrived at this point, password or ssh_keyfile
        # is set. Remove the not set one
        if 'password' in cred_keys:
            cred_keys.remove('password')

        if 'ssh_keyfile' in cred_keys:
            cred_keys.remove('ssh_keyfile')

        cred_keys = [x for x in cred_keys
                     if x not in self._optional_cred_keys]
        return list(cred_keys)



from pygments.token import Token

from nubia import statusbar
from pandas import DataFrame

from suzieq.version import SUZIEQ_VERSION


class NubiaSuzieqStatusBar(statusbar.StatusBar):
    '''Process CLI statusbar updates'''

    def __init__(self, ctxt):
        self._last_status = None
        self.ctxt = ctxt

    def get_rprompt_tokens(self):
        if not isinstance(self._last_status, DataFrame) and self._last_status:
            return [(Token.RPrompt, "Error: {}".format(self._last_status))]
        return []

    def set_last_command_status(self, status):
        self._last_status = status

    def get_tokens(self):
        spacer = (Token.Spacer, "  ")
        ctxt = self.ctxt.ctxt
        if ctxt.pager:
            is_pager = (Token.Warn, "ON")
        else:
            is_pager = (Token.Info, "OFF")

        if ctxt.engine != "rest":
            engine = ctxt.engine
        else:
            engine = (f'{ctxt.rest_server_ip}:'
                      f'{ctxt.rest_server_port}')
        return [
            (Token.Toolbar, "Suzieq"),
            spacer,
            (Token.Toolbar, "Version "),
            spacer,
            (Token.Info, SUZIEQ_VERSION),
            spacer,
            (Token.Toolbar, "Pager "),
            spacer,
            is_pager,
            spacer,
            (Token.Toolbar, "Namespace "),
            spacer,
            (Token.Info, ", ".join(ctxt.namespace)),
            spacer,
            (Token.Toolbar, "Hostname "),
            spacer,
            (Token.Info, ", ".join(ctxt.hostname)),
            spacer,
            (Token.Toolbar, "StartTime "),
            spacer,
            (Token.Info, ctxt.start_time),
            spacer,
            (Token.Toolbar, "EndTime "),
            spacer,
            (Token.Info, ctxt.end_time),
            spacer,
            (Token.Toolbar, "Engine "),
            spacer,
            (Token.Info, engine),
            spacer,
            (Token.Toolbar, "Query Time "),
            spacer,
            (Token.Info, ctxt.exec_time),
        ]



from suzieq.sqobjects.basicobj import SqObject


class PathObj(SqObject):
    '''The object providing access to the path table'''

    def __init__(self, **kwargs):
        super().__init__(table='path', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns',
                                'vrf', 'src', 'dest', 'query_str']
        self._valid_summarize_args = ['namespace', 'hostname', 'src',
                                      'dest', 'vrf', 'query_str']



from typing import List
import re
from ipaddress import ip_address, ip_network

import dateparser
import numpy as np
import pandas as pd
from pandas.core.groupby import DataFrameGroupBy

from suzieq.shared.utils import humanize_timestamp
from suzieq.shared.schema import Schema, SchemaForTable
from suzieq.engines.base_engine import SqEngineObj
from suzieq.sqobjects import get_sqobject
from suzieq.db import get_sqdb_engine
from suzieq.shared.exceptions import UserQueryError


class SqPandasEngine(SqEngineObj):
    '''Base class to handle manipulating tables with pandas'''

    def __init__(self, baseobj):
        self.ctxt = baseobj.ctxt
        self.iobj = baseobj
        self.summary_row_order = []
        self.nsgrp = None
        self.ns_df = pd.DataFrame()
        self.ns = []
        self.summary_df = pd.DataFrame()
        self._summarize_on_add_field = []
        self._summarize_on_add_with_query = []
        self._summarize_on_add_list_or_count = []
        self._summarize_on_add_stat = []
        self._summarize_on_perdevice_stat = []
        self._check_empty_col = 'namespace'
        self._dbeng = get_sqdb_engine(baseobj.ctxt.cfg, baseobj.table, '',
                                      None)

    @property
    def all_schemas(self) -> Schema:
        '''Return dict of all schemas'''
        return self.ctxt.schemas

    @property
    def cfg(self):
        '''Return config dicttionary'''
        return self.iobj.cfg

    @property
    def name(self):
        return 'pandas'

    @property
    def table(self):
        '''Table name'''
        return self.iobj.table

    @property
    def schema(self) -> SchemaForTable:
        '''Return schema for this table'''
        return self.iobj.schema

    def _get_ipvers(self, value: str) -> int:
        """Return the IP version in use"""

        if ':' in value:
            ipvers = 6
        elif '.' in value:
            ipvers = 4
        else:
            ipvers = ''

        return ipvers

    def _handle_user_query_str(self, df: pd.DataFrame,
                               query_str: str) -> pd.DataFrame:
        """Handle user query, trapping errors and returning exception

        Args:
            df (pd.DataFrame): The dataframe to run the query on
            query_str (str): pandas query string

        Raises:
            UserQueryError: Exception if pandas query aborts with errmsg

        Returns:
            pd.DataFrame: dataframe post query
        """
        if query_str:
            try:
                df = df.query(query_str)
            except Exception as ex:
                raise UserQueryError(ex) from Exception

        return df

    def _is_any_in_list(self, column: pd.Series, values: List) -> pd.Series:
        """Given a column of arrays, check if any of the values is present in
        the list.
        Args:
            column (pd.Series): pandas dataframe column of lists
            values (List): list of values representing the filtering values
        Returns:
            pd.Series: A collection of bool reporting the filtered result
        """
        return column.apply(lambda x: any(v in x for v in values))

    def _is_in_subnet(self, addr: pd.Series, net: str) -> pd.Series:
        """Check if the IP addresses in a Pandas dataframe
        belongs to the given subnet

        Args:
            addr (PandasObject): the collection of ip addresses to check
            net: (str): network id of the subnet

        Returns:
            PandasObject: A collection of bool reporting the result
        """
        network = ip_network(net)
        if isinstance(addr.iloc[0], np.ndarray):
            return addr.apply(lambda x, network:
                              False if not x.any()
                              else any(ip_address(a.split("/")[0]) in network
                                       for a in x),
                              args=(network,))
        else:
            return addr.apply(lambda a: (
                False if not a else ip_address(a.split("/")[0]) in network)
            )

    def _check_ipvers(self,  addr: pd.Series, version: int) -> pd.Series:
        """Check if the IP version of addresses in a Pandas dataframe
        correspond to the given version

        Args:
            addr (PandasObject): the collection of ip addresses to check
            version: (int): IP version (4 or 6)

        Returns:
            PandasObject: A collection of bool reporting the result
        """

        return addr.apply(lambda a: (self._get_ipvers(a) == version))

    # pylint: disable=too-many-statements
    def get_valid_df(self, **kwargs) -> pd.DataFrame:
        """The heart of the engine: retrieving the data from the backing store

        Args:
            kwargs: keyword args passed by caller, varies depending on table

        Returns:
            pd.DataFrame: The data as a pandas dataframe
        """
        if not self.ctxt.engine:
            print("Specify an analysis engine using set engine command")
            return pd.DataFrame(columns=["namespace", "hostname"])

        # Thanks to things like OSPF, we cannot use self.schema here
        sch = self.schema
        phy_table = sch.get_phy_table_for_table()

        columns = kwargs.pop('columns', ['default'])
        addnl_fields = kwargs.pop('addnl_fields', [])
        view = kwargs.pop('view', self.iobj.view)
        # active_only tells the db to get only the active records (discarding
        # everything have been deleted).
        # If view=all, by default, we would like to show all the records,
        # included the ones that have been deleted, in all the other cases
        # we want to see only the active records.
        active_only = kwargs.pop('active_only', view != 'all')
        hostname = kwargs.pop('hostname', [])
        query_str = kwargs.pop('query_str', '')

        fields = sch.get_display_fields(columns)
        key_fields = sch.key_fields()

        # Lets also get the fields required to satisfy user-specified filters
        # We cannot use sets because we want to preserve the order of fields
        filter_fields = [x for x in self.iobj.get_filter_fields
                         if x not in addnl_fields]
        user_query_cols = [x for x in self._get_user_query_cols(query_str)
                           if x not in addnl_fields]
        fields += [x for x in user_query_cols+filter_fields+addnl_fields
                   if x not in fields]

        aug_fields = sch.get_augmented_fields(fields)

        if 'timestamp' not in fields:
            fields.append('timestamp')

        self._add_active_to_fields(view, fields, addnl_fields)
        # Order matters. Don't put this before the missing key fields insert
        for f in aug_fields:
            dep_fields = sch.get_parent_fields(f)
            dep_fields = [x for x in dep_fields if x not in aug_fields and
                          x not in fields]
            fields += dep_fields
            addnl_fields += dep_fields

        # Remove augmented fields from being passed to DB read.
        # Augmented fields are computed from other fields
        fields = [x for x in fields if x not in aug_fields]

        for fld in key_fields:
            if fld not in fields+addnl_fields:
                addnl_fields.insert(0, fld)

        getcols = list(set(fields+addnl_fields))

        tz = self.ctxt.cfg.get('analyzer', {}).get('timezone', 'UTC')
        settings = {'TIMEZONE': tz}

        if self.iobj.start_time:
            try:
                start_time = int(dateparser.parse(
                    self.iobj.start_time.replace('last night', 'yesterday'),
                    settings=settings)
                    .timestamp()*1000)
            except Exception:
                # pylint disable=raise-missing-from
                raise ValueError(
                    f"unable to parse start-time: {self.iobj.start_time}")
        else:
            start_time = ''

        if self.iobj.start_time and not start_time:
            # Something went wrong with our parsing
            # pylint disable=raise-missing-from
            raise ValueError(
                f"unable to parse start-time: {self.iobj.start_time}")

        if self.iobj.end_time:
            try:
                end_time = int(dateparser.parse(
                    self.iobj.end_time.replace('last night', 'yesterday'),
                    settings=settings)
                    .timestamp()*1000)
            except Exception:
                # pylint disable=raise-missing-from
                raise ValueError(
                    f"unable to parse end-time: {self.iobj.end_time}")
        else:
            end_time = ''

        if self.iobj.end_time and not end_time:
            # Something went wrong with our parsing
            # pylint disable=raise-missing-from
            raise ValueError(
                f"unable to parse end-time: {self.iobj.end_time}")

        table_df = self._dbeng.read(
            phy_table,
            'pandas',
            start_time=start_time,
            end_time=end_time,
            columns=getcols,
            view=view,
            key_fields=key_fields,
            **kwargs
        )

        if not table_df.empty:
            # hostname may not have been filtered if using regex
            if hostname:
                table_df = self._filter_hostname(table_df, hostname)
            if active_only:
                table_df = table_df.query('active').reset_index(drop=True)

            if 'timestamp' in table_df.columns and not table_df.empty:
                table_df['timestamp'] = humanize_timestamp(
                    table_df.timestamp, self.cfg.get('analyzer', {})
                    .get('timezone', None))

            if query_str:
                table_df = self._handle_user_query_str(table_df, query_str)

        return table_df.reset_index(drop=True)[fields]

    def _filter_hostname(self, df: pd.DataFrame, hostname: List[str]) \
            -> pd.DataFrame:
        """filter dataframe with the input hostnames

        Args:
            hostname (List[str]): hostname used for filter df
            df (pd.Dataframe): input dataframe

        Returns:
            pd.DataFrame: filtered dataframe
        """
        if hostname and not df.empty:
            hdf_list = []
            notlist = [x for x in hostname if '!' in x]
            if len(notlist) != len(hostname):
                hnlist = [x for x in hostname if '!' not in x]
            else:
                hnlist = hostname
            for hn in hnlist:
                use_not = False
                if hn.startswith('~'):
                    hn = hn[1:]
                    if hn.startswith('!'):
                        use_not = True
                        hn = hn[1:]
                elif hn.startswith('!'):
                    hn = hn[1:]
                    use_not = True

                if use_not:
                    df1 = df.query(f'~hostname.str.match("{hn}")')
                    hdf_list = [df1]
                else:
                    df1 = df.query(f"hostname.str.match('{hn}')")
                    if not df1.empty:
                        hdf_list.append(df1)

                if use_not:
                    # With not, the list of hostnames becomes an and
                    df = df1

            if hdf_list:
                df = pd.concat(hdf_list)
            else:
                return pd.DataFrame(columns=df.columns.tolist())
        return df

    def get(self, **kwargs) -> pd.DataFrame:
        """The default get method for all tables

        Use this for a table if nothing special is desired. No table uses
        this routine today.

        Raises:
            NotImplementedError: If no table has been defined

        Returns:
            pd.DataFrame: pandas dataframe of the object
        """
        if not self.iobj.table:
            raise NotImplementedError

        df = self.get_valid_df(**kwargs)

        return df

    def get_table_info(self, **kwargs) -> dict:
        """Returns information about the data available for a table

        Used by table show command exclusively.
        Args:
            kwargs: keyword args passed by caller, varies depending on table

        Returns:
            dict: The desired data as a dictionary
        """
        # You can't use view from user because we need to see all the data
        # to compute data required.
        kwargs.pop('view', None)

        all_time_df = self.get_valid_df(view='all', **kwargs)
        times = all_time_df['timestamp'].unique()
        ret = {'firstTime': all_time_df.timestamp.min(),
               'latestTime': all_time_df.timestamp.max(),
               'intervals': len(times),
               'allRows': len(all_time_df),
               'namespaces': all_time_df.get('namespace',
                                             pd.Series(dtype='category'))
               .nunique(),
               'deviceCnt': all_time_df.get('hostname',
                                            pd.Series(dtype='category'))
               .nunique(),
               }
        return ret

    def summarize(self, **kwargs):
        """Summarize the info about this resource/service.

        There is a pattern of how to do these
        use self._init_summarize():
           - creates self.summary_df, which is the initial pandas dataframe
             based on the table
           - creates self.nsgrp of data grouped by namespace
           - self.ns is the dict to add data to which will be turned into a
             dataframe and then returned

        if you want to simply take a field and run a pandas functon, then use
          self._add_field_to_summary

        at the end of te summarize
        return pd.DataFrame(self.ns).convert_dtypes()

        If you don't override this, then you get a default summary of all
        columns
        """
        self._init_summarize(**kwargs)
        if self.summary_df.empty:
            return self.summary_df

        self._gen_summarize_data()
        self._post_summarize()
        return self.ns_df.convert_dtypes()

    def unique(self, **kwargs) -> pd.DataFrame:
        """Return the unique elements as per user specification

        Raises:
            ValueError: If len(columns) != 1

        Returns:
            pd.DataFrame: Pandas dataframe of unique values for given column
        """
        count = kwargs.pop("count", 'False') == 'True'
        query_str = kwargs.pop('query_str', '')
        get_query_str = kwargs.pop('get_query_str', '')

        columns = kwargs.pop("columns", None)

        column = columns[0]

        df = self.get(columns=columns, query_str=get_query_str, **kwargs)
        if df.empty:
            return df

        # check if column we're looking at is a list, and if so explode it
        if df[column].apply(
                lambda x: isinstance(x, (list, np.ndarray))).any():
            df = df.explode(column).dropna(how='any')

        if count:
            r = df[column].value_counts()
            df = pd.DataFrame({column: r}) \
                .reset_index() \
                .rename(columns={column: 'numRows',
                                 'index': column}) \
                   .sort_values(column)
        else:
            df = pd.DataFrame({f'{column}': df[column].unique()})

        if query_str:
            df = self._handle_user_query_str(df, query_str)

        return df

    def aver(self, **kwargs):
        '''Assert'''
        raise NotImplementedError

    def top(self, **kwargs):
        """Default implementation of top.
        The basic fields this assumes are present include the "what" keyword
        which contains the name of the field we're getting the transitions on,
        the "n" field which tells the count of the top entries you're
        looking for, and the reverse field which tells whether you're looking
        for the largest (default, and so reverse is False) or the smallest(
        reverse is True). This invokes the default object's get routine. It
        is upto the caller to ensure that the desired column is in the output.
        """
        what = kwargs.pop("what", None)
        reverse = kwargs.pop("reverse", False)
        sqTopCount = kwargs.pop("count", 5)
        columns = kwargs.pop("columns", ['default'])

        if not what:
            return pd.DataFrame()

        columns = self.schema.get_display_fields(columns)
        if what not in columns:
            columns.insert(-1, what)

        df = self.get(columns=columns, **kwargs)
        if df.empty or ('error' in df.columns):
            return df

        columns_by = [what] + self.schema.key_fields()

        return df.sort_values(by=columns_by, ascending=reverse) \
                 .head(sqTopCount)

    def lpm(self, **kwargs):
        '''Default implementation to return not supported'''
        return pd.Dataframe(
            {'error': [f'Unsupported verb "lpm" for {self.iobj.table}']})

    def find(self, **kwargs):
        '''Default implementation to return not supported'''
        return pd.Dataframe(
            {'error': [f'Unsupported verb "find" for {self.iobj.table}']})

    def _get_table_sqobj(self, table: str, start_time: str = None,
                         end_time: str = None, view=None):
        """Normalize pulling data from other tables into this one function

        Typically pulling data involves calling get_sqobject with a bunch of
        parameters that need to be passed to it, that a caller can forget to
        pass. A classic example is passing the view, start-time and end-time
        which is often forgotten. This function fixes this issue.

        Args:
            table (str): The table to retrieve the info from
            verb (str): The verb to use in the get_sqobject call
        """

        return get_sqobject(table)(
            engine_name=self.iobj.engine.name,
            context=self.ctxt,
            start_time=start_time or self.iobj.start_time,
            end_time=end_time or self.iobj.end_time,
            view=view or self.iobj.view)

    def _init_summarize(self, **kwargs) -> None:
        """Initialize the data structures for use with generating summary

        """
        kwargs.pop('columns', None)
        columns = ['*']

        df = self.get(columns=columns, **kwargs)
        if 'error' in df.columns:
            self.summary_df = df
            return

        self.summary_df = df
        if df.empty:
            return

        self.ns = {i: {} for i in df['namespace'].unique()}
        self.nsgrp = df.groupby(by=["namespace"], observed=True)

    def _gen_summarize_data(self):
        """Generate the data required for summary"""

        if not self._summarize_on_add_field:
            # Add the only field we truly know to add
            self._summarize_on_add_field = [
                ('deviceCnt', 'hostname', 'nunique'),
            ]
        for field_name, col, function in self._summarize_on_add_field:
            if col not in ['namespace', 'timestamp']:
                self._add_field_to_summary(col, function, field_name)
                self.summary_row_order.append(field_name)

        for flds in self._summarize_on_add_with_query:
            field_name = flds[0]
            query_str = flds[1]
            field = flds[2]
            if len(flds) == 4:
                func = flds[3]
            else:
                func = 'count'
            fld_df = self.summary_df.query(query_str)
            if not fld_df.empty:
                fld_per_ns = fld_df.groupby(by=['namespace'],
                                            observed=True)[field] \
                    .agg(func)
                for i in self.ns.keys():
                    self.ns[i].update({field_name: fld_per_ns.get(i, 0)})
            else:
                for i in self.ns.keys():
                    self.ns[i].update({field_name: 0})

            self.summary_row_order.append(field_name)

        for field_name, field in self._summarize_on_add_list_or_count:
            self._add_list_or_count_to_summary(field, field_name)
            self.summary_row_order.append(field_name)

        for field_name, query_str, field in self._summarize_on_add_stat:
            if query_str:
                statfld = self.summary_df.query(query_str) \
                    .groupby(by=['namespace'],
                             observed=True)[field]
            else:
                statfld = self.summary_df.groupby(
                    by=['namespace'], observed=True)[field]

            self._add_stats_to_summary(statfld, field_name)
            self.summary_row_order.append(field_name)

        for field_name, query_str, field, func in \
                self._summarize_on_perdevice_stat:
            if query_str:
                statfld = self.summary_df \
                    .query(query_str) \
                    .groupby(by=['namespace', 'hostname'],
                             observed=True)[field].agg(func)
            else:
                statfld = self.summary_df \
                    .groupby(by=['namespace', 'hostname'],
                             observed=True)[field].agg(func)

            self._add_stats_to_summary(statfld, field_name, filter_by_ns=True)
            self.summary_row_order.append(field_name)

    def _post_summarize(self, check_empty_col: str = 'deviceCnt') -> None:
        """Once summary data has been generated, finalize summary dataframe

        Args:
            check_empty_col (str, optional): column name to check to remove
                                             namespace that's empty.
                                             Defaults to 'deviceCnt'.
        """
        # this is needed in the case that there is a namespace that has no
        # data for this command

        if not check_empty_col:
            check_empty_col = self._check_empty_col

        delete_keys = []
        for ns in self.ns:
            if self.ns[ns][check_empty_col] == 0:
                delete_keys.append(ns)
        for ns in delete_keys:
            del(self.ns[ns])

        ns_df = pd.DataFrame(self.ns)
        if len(self.summary_row_order) > 0:
            ns_df = ns_df.reindex(self.summary_row_order, axis=0)
        self.ns_df = ns_df

    def _add_constant_to_summary(self, field: str, value):
        """And a constant value to specified field name in summary"""
        if field:
            # pylint disable=expression-not-assigned
            _ = {self.ns[i].update({field: value}) for i in self.ns.keys()}

    def _add_field_to_summary(self, field: str, method: str = 'nunique',
                              field_name: str = None):
        """Add a summary field to summary dataframe based on method on field

        This assumes that self.summary_df has been populated.
        Args:
            field (str): Column name to add in the summary dataframe
            method (str, optional): pandas method name. Defaults to 'nunique'.
            field_name (str, optional): column to compute on. Defaults to None.
        """
        if not field_name:
            field_name = field
        field_per_ns = getattr(self.nsgrp[field], method)()
        _ = {self.ns[i].update({field_name: field_per_ns.get(i, 0)})
             for i in self.ns.keys()}

    def _add_list_or_count_to_summary(self, field: str,
                                      field_name: str = None) -> None:
        """add as a list if < 3 things, otherwise return the count"""
        if not field_name:
            field_name = field

        count_per_ns = self.nsgrp[field].nunique()

        for n in self.ns.keys():
            if 3 >= count_per_ns[n] > 0:
                # can't do a value_counts on all groups, incase one of the
                # groups other groups doesn't have data
                unique_for_ns = self.nsgrp.get_group(n)[field].value_counts()
                value = unique_for_ns.to_dict()
                # Filter numm entries if category because of how pandas
                # behaves here
                if self.nsgrp[field].dtype[n].name == 'category':
                    value = dict(filter(lambda x: x[1] != 0, value.items()))

            else:
                value = count_per_ns[n]

            self.ns[n].update({field_name: value})

    def _add_stats_to_summary(self, groupedby: DataFrameGroupBy,
                              fieldname: str,
                              filter_by_ns: bool = False) -> None:
        """Takes grouped stats and adds min, max, and median to stats"""

        _ = {self.ns[i].update({fieldname: []}) for i in self.ns.keys()}
        if filter_by_ns:
            _ = {self.ns[i][fieldname].append(groupedby[i].min())
                 if i in groupedby else self.ns[i][fieldname].append(0)
                 for i in self.ns.keys()}
            # pylint disable=expression-not-assigned
            _ = {self.ns[i][fieldname].append(groupedby[i].max())
                 if i in groupedby else self.ns[i][fieldname].append(0)
                 for i in self.ns.keys()}
            # pylint disable=expression-not-assigned
            _ = {self.ns[i][fieldname].append(
                groupedby[i].median(numeric_only=False))
                if i in groupedby else self.ns[i][fieldname].append(0)
                for i in self.ns.keys()}
        else:
            min_field = groupedby.min()
            max_field = groupedby.max()
            med_field = groupedby.median(numeric_only=False)

            _ = {self.ns[i][fieldname].append(min_field[i])
                 if i in min_field else self.ns[i][fieldname].append(0)
                 for i in self.ns.keys()}
            _ = {self.ns[i][fieldname].append(max_field[i])
                 if i in max_field else self.ns[i][fieldname].append(0)
                 for i in self.ns.keys()}
            _ = {self.ns[i][fieldname].append(med_field[i])
                 if i in med_field else self.ns[i][fieldname].append(0)
                 for i in self.ns.keys()}

    def _get_user_query_cols(self, query_str: str) -> List[str]:
        """Returns the list of fields inside the query_str

        Args:
            query_str (str): query string
            table (str): table from which get the fields

        Returns:
            List[str]: list of fields inside the query string
        """
        if query_str:
            table_fields = self.schema.fields
            return [f for f in table_fields if re.search(rf'\b{f}\b',
                                                         query_str)]
        return []

    def _add_active_to_fields(self, view: str, fields: List[str],
                              addnl_fields: List[str]) -> None:
        """Add 'active' field to returned cols

        Depending on the user specification, if we have to return all
        records within a time window or if view=all, we must also
        return the active field because it marks if the record was
        added or deleted

        Args:
            view(str): value of view specified by user
            fields(List[str]): the columns to be returned so far
            addnl_fields(List[str]): the addnl fields to be retrieved so far

        Returns:
            None
        """

        if view == "all" or (self.iobj.start_time and self.iobj.end_time):
            if 'active' not in fields:
                fields.insert(0, 'active')
        elif addnl_fields is not None:
            addnl_fields.append('active')



import time
from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqTableCommand
from suzieq.sqobjects.namespace import NamespaceObj


@command("namespace", help="Summarize namespace-wide network data")
class NamespaceCmd(SqTableCommand):
    """Overall network information such as device count, bgp enabled etc."""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            query_str: str = " ",
            columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=NamespaceObj,
        )

    @command("show", help="Show namespace information")
    @argument("model", description="Device model(s), space separated")
    @argument("os", description='Device NOS(es), space separated')
    @argument('vendor', description='Device vendor(s), space separated')
    @argument('version', description='Device NOS version(s), space separated')
    def show(self, os: str = "", vendor: str = "", model: str = "",
             version: str = "") -> int:
        """Show namespace info
        """

        now = time.time()

        df = self._invoke_sqobj(self.sqobj.get,
                                namespace=self.namespace, os=os.split(),
                                vendor=vendor.split(), model=model.split(),
                                version=version, query_str=self.query_str,
                                hostname=self.hostname)

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._gen_output(df)



from dataclasses import dataclass, field
from typing import Any, List

import altair as alt
import numpy as np
import pandas as pd
import streamlit as st
from st_aggrid import AgGrid, GridOptionsBuilder, GridUpdateMode, JsCode
from suzieq.gui.stlit.guiutils import (SUZIEQ_COLOR, SuzieqMainPages,
                                       gui_get_df, pandas_df_to_markdown_table,
                                       set_def_aggrid_options, sq_gui_style)
from suzieq.gui.stlit.pagecls import SqGuiPage
from suzieq.sqobjects import get_sqobject, get_tables

DEF_MAX_ROW = 10000
MAX_ROW_WARN = 50000


@dataclass
class XploreSessionState:
    '''Session state for Xplore page'''
    page: str = SuzieqMainPages.XPLORE.value
    namespace: str = ''
    start_time: str = ''
    end_time: str = ''
    table: str = ''
    view: str = 'latest'
    query: str = ''
    columns:  List[str] = field(default_factory=list)
    uniq_clicked: str = '-'
    assert_clicked: bool = False
    get_clicked: bool = False
    col_sel_done: bool = False
    experimental_ok: bool = True
    start_row: int = 0
    end_row: int = DEF_MAX_ROW
    tables_obj: Any = None


class XplorePage(SqGuiPage):
    '''Page for exploratory analysis'''
    _title: str = SuzieqMainPages.XPLORE.value
    _state: XploreSessionState = XploreSessionState()
    _reload_data: bool = False
    _config_file: str = st.session_state.get('config_file', '')
    _cached_df: pd.DataFrame = pd.DataFrame()

    @property
    def add_to_menu(self) -> bool:
        return True

    def build(self):
        self._get_state_from_url()
        self._create_sidebar()
        layout = self._create_layout()
        self._render(layout)

    # pylint: disable=too-many-statements
    def _create_sidebar(self) -> None:

        state = self._state
        stime = state.start_time
        etime = state.end_time

        tables = filter(
            lambda x: x not in ['path', 'tables', 'ospfIf', 'ospfNbr',
                                'devconfig', 'topmem', 'topcpu', 'ifCounters',
                                'time', 'network'],
            get_tables()
        )
        table_vals = [''] + sorted(tables)
        if state.table and state.table not in table_vals:
            st.error(f'Unknown table {state.table}, ignoring')
            state.table = ''

        if state.table:
            if isinstance(state.table, list):
                tblidx = table_vals.index(state.table[0])
            else:
                tblidx = table_vals.index(state.table)
        else:
            tblidx = table_vals.index('namespace')  # Default starting table
        view_idx = 1 if state.view == 'all' else 0

        devdf = gui_get_df('device', self._config_file,
                           columns=['namespace', 'hostname'])
        if devdf.empty:
            st.error('Unable to retrieve any namespace info')
            st.stop()

        namespaces = [""]
        namespaces.extend(sorted(devdf.namespace.unique().tolist()))
        if state.namespace:
            nsidx = namespaces.index(state.namespace)
        else:
            nsidx = 0
        with st.sidebar:
            with st.form('Xplore'):
                namespace = st.selectbox('Namespace',
                                         namespaces, key='xplore_namespace',
                                         index=nsidx)
                state.start_time = st.text_input('Start time',
                                                 value=stime,
                                                 key='xplore_stime')
                state.end_time = st.text_input('End time',
                                               value=etime,
                                               key='xplore_etime')

                table = st.selectbox(
                    'Select Table to View', tuple(table_vals),
                    key='xplore_table', index=tblidx)
                if table != state.table:
                    # We need to reset the specific variables
                    state.query = ''
                    state.assert_clicked = False
                    state.uniq_clicked = '-'
                    state.columns = ['default']
                    state.table = table

                view_vals = ('latest', 'all')
                if state.start_time and state.end_time:
                    # Show everything thats happened if both times are given
                    view_idx = 1
                state.view = st.radio("View of Data", view_vals,
                                      index=view_idx, key='xplore_view')
                st.form_submit_button('Get', on_click=self._fetch_data)

            if state.table:
                state.tables_obj = get_sqobject(state.table)()
                fields = state.tables_obj.describe()
                colist = sorted((filter(lambda x: x not in ['index', 'sqvers'],
                                        fields.name.tolist())))
                colist = ['default', 'all'] + colist
                # remove invalid columns from the state
                state.columns = [c for c in state.columns if c in colist]
                with st.form('Columns'):
                    state.columns = st.multiselect(
                        'Pick columns', colist,
                        key='xplore_columns', default=state.columns)
                    st.form_submit_button('Selection Done',
                                          on_click=self._update_cols)

            state.start_row = st.number_input('Start Row Index',
                                              value=int(state.start_row),
                                              min_value=0,
                                              step=1,
                                              key='xplore_start_row',
                                              on_change=self._update_row_sel)
            state.start_row = int(state.start_row)

            state.end_row = st.number_input('End Row Index',
                                            value=int(state.end_row),
                                            key='xplore_end_row',
                                            step=1,
                                            on_change=self._update_row_sel)
            state.end_row = int(state.end_row)

        col_ok = True
        if ((('default' in state.columns) or ('*' in state.columns))
                and len(state.columns) != 1):
            self._save_page_url()
            st.info('You cannot select default/all with other columns')
            st.stop()

        if not state.columns:
            self._save_page_url()
            st.info('Please select some columns to continue')
            st.stop()

        if namespace != state.namespace:
            state.namespace = namespace

        if state.columns == ['all']:
            state.columns = ['*']
        columns = state.columns

        state.experimental_ok = st.sidebar.checkbox(
            'Enable Experimental Features',
            value=state.experimental_ok,
            key='xplore_exp',
            on_change=self._sync_state)
        if state.table in ['interfaces', 'ospf', 'bgp', 'evpnVni']:
            state.assert_clicked = st.sidebar.checkbox(
                'Run Assert',  key='xplore_assert',
                on_change=self._sync_state)
        else:
            state.assert_clicked = False

        state.query = st.sidebar.text_input(
            'Filter results with pandas query', value=state.query,
            key='xplore_query', on_change=self._sync_state)

        st.sidebar.markdown(
            "[query syntax help]"
            "(https://suzieq.readthedocs.io/en/latest/pandas-query-examples/)")

        if state.get_clicked:
            state.get_clicked = False
        elif not col_ok and not self._first_time:
            self._save_page_url()
            st.info("Click 'Get' button to show the data")
            st.stop()
        if ('default' in columns or 'all' in columns) and len(columns) != 1:
            st.error('Cannot select default/all with any other columns')
            st.stop()
        elif not columns:
            st.error('Columns cannot be empty')
            st.stop()
        if self._first_time:
            self._first_time = False

    def _create_layout(self) -> None:
        grid1 = st.container()
        with grid1:
            headercol, uniq_col = st.columns(2)
            header_ph = headercol.empty()
            uniq_ph = uniq_col.container()

            scol1, scol2 = st.columns(2)
            scol1_ph = scol1.empty()
            scol2_ph = scol2.empty()
            if self._state.table in ['interfaces', 'ospf', 'bgp', 'evpnVni']:
                validate_ph = st.empty()
            else:
                validate_ph = None
            table_ph = st.empty()
            col_names_ph = st.empty()

        return {
            'header': header_ph,  # summary header
            'uniq_col': uniq_ph,  # unique column selector
            'summary': scol1_ph,  # summary table
            'uniq': scol2_ph,  # unique table
            'assert': validate_ph,  # assert expander
            'table': table_ph,  # table dataframe expander
            'col_names': col_names_ph  # column name table expander
        }

    def _render(self, layout: dict) -> None:

        state = self._state

        if not state.table:
            return
        query_str = state.query
        try:
            df = gui_get_df(state.table,
                            self._config_file,
                            namespace=state.namespace.split(),
                            start_time=state.start_time,
                            end_time=state.end_time,
                            view=state.view, columns=state.columns,
                            query_str=query_str)
        except Exception as e:  # pylint: disable=broad-except
            st.error(e)
            st.stop()

        if not df.empty:
            if 'error' in df.columns:
                st.error(df.iloc[0].error)
                self._save_page_url()
                st.stop()
                return
        else:
            st.info('No data returned by the table')
            self._save_page_url()
            st.stop()
            return

        if not df.empty:
            self._draw_summary_df(layout, query_str)
            self._draw_assert_df(layout)

        self._draw_table_df(layout, df)

    def _sync_state(self) -> None:
        wsstate = st.session_state
        state = self._state
        change_view = False

        if wsstate.xplore_namespace != state.namespace:
            state.namespace = wsstate.xplore_namespace

        if ((wsstate.xplore_stime != state.start_time) or
                (wsstate.xplore_etime != state.end_time)):
            if not (wsstate.xplore_stime and wsstate.xplore_etime):
                change_view = True
        if wsstate.xplore_stime != state.start_time:
            state.start_time = wsstate.xplore_stime
        if wsstate.xplore_etime != state.end_time:
            state.end_time = wsstate.xplore_etime
        if wsstate.xplore_view != state.view:
            state.view = wsstate.xplore_view
        if change_view:
            state.view = 'latest'
        if wsstate.xplore_query != state.query:
            state.query = wsstate.xplore_query
        assert_clicked = wsstate.get('xplore_assert', '')
        if assert_clicked and (assert_clicked != state.assert_clicked):
            state.assert_clicked = wsstate.xplore_assert
        if wsstate.xplore_exp != state.experimental_ok:
            state.experimental_ok = wsstate.xplore_exp

        if wsstate.xplore_table != state.table:
            # Reset all dependent vars when table changes
            if f'xplore_show_{state.table}' in wsstate:
                wsstate[f'xplore_show_{state.table}'] = pd.DataFrame()
            state.table = wsstate.xplore_table
            state.query = ''
            state.assert_clicked = False
            state.uniq_clicked = '-'
            state.columns = ['default']
            state.start_row = 0
            state.end_row = DEF_MAX_ROW
        self._save_page_url()

    def _update_cols(self):
        wsstate = st.session_state
        state = self._state

        state.columns = wsstate.xplore_columns
        self._save_page_url()

    def _update_row_sel(self):
        wsstate = st.session_state
        state = self._state

        if int(wsstate.xplore_start_row) != state.start_row:
            state.start_row = int(wsstate.xplore_start_row)

        if int(wsstate.xplore_end_row) != state.end_row:
            state.end_row = int(wsstate.xplore_end_row)

        self._save_page_url()

    def _fetch_data(self):
        '''Called when the Get button is pressed, refresh data'''
        table = getattr(self._state, 'table', 'namespace')
        if f'xplore_show_{table}' in st.session_state:
            del st.session_state[f'xplore_show_{table}']
        self._state.get_clicked = True
        self._sync_state()

    def _sync_uniq_col(self):
        state = self._state
        uniq_sel = st.session_state.get(f'xplore_uniq_col_{state.table}',
                                        '')
        if uniq_sel and (state.uniq_clicked != uniq_sel):
            state.uniq_clicked = uniq_sel

    def _draw_table_df(self, layout, show_df):
        '''Display the table dataframe'''

        if not show_df.empty:
            agdf = show_df
            ag = self._draw_aggrid_df(agdf)
            selected = ag['data']
            selected_df = pd.DataFrame(selected)
            st.session_state['xplore_show_df'] = selected_df
            self._draw_uniq_histogram(layout, selected_df)

            help_df = self._state.tables_obj \
                .describe() \
                .query('name != "sqvers"') \
                .drop(['key', 'display'], axis=1) \
                .reset_index(drop=True)

            help_table = pandas_df_to_markdown_table(help_df)
            st.button("?", help=help_table)
        else:
            st.warning('No Data from query')

    def _draw_aggrid_df(self, df) -> AgGrid:

        gb = GridOptionsBuilder.from_dataframe(df)
        gb.configure_pagination(paginationPageSize=25)

        gb.configure_default_column(floatingFilter=True, selectable=False)

        gb.configure_grid_options(domLayout='normal')

        gridOptions = gb.build()
        gridOptions = set_def_aggrid_options(gridOptions)
        jscode = self._aggrid_style_rows()
        gridOptions['getRowStyle'] = jscode

        start_row = self._state.start_row
        end_row = self._state.end_row
        if end_row > df.shape[0]:
            end_row = df.shape[0]

        if end_row - start_row > MAX_ROW_WARN:
            st.warning('Printing too many rows can fail or be very slow')

        if (end_row - start_row) < df.shape[0]:
            st.info(f'Displaying rows {start_row} to {end_row} of '
                    f'{df.shape[0]} rows')

        if self._state.experimental_ok:
            retmode = 'FILTERED'
            upd8_mode = GridUpdateMode.FILTERING_CHANGED
        else:
            retmode = 'AS_INPUT'
            upd8_mode = GridUpdateMode.VALUE_CHANGED

        fit_columns = (len(df.columns) < 12)
        grid_response = AgGrid(
            df.iloc[start_row:end_row],
            gridOptions=gridOptions,
            allow_unsafe_jscode=True,
            data_return_mode=retmode,
            update_mode=upd8_mode,
            fit_columns_on_grid_load=fit_columns,
            theme='streamlit',
        )
        return grid_response

    def _aggrid_style_rows(self):
        '''Style the cells based on value'''
        table = self._state.table

        style_dict = {
            'bgp': """
                   function(params) {
                      if (params.data.state === 'NotEstd') {
                         return {
                            'color': 'white',
                            'backgroundColor': 'red'
                         }
                      }
                   };
                   """,
            'device': """
                        function(params) {
                            if (params.data.status === 'dead') {
                                return {
                                    'color': 'white',
                                    'backgroundColor': 'red'
                                }
                            } else if (params.data.status === 'neverpoll') {
                                return {
                                    'color': 'white',
                                    'backgroundColor': 'darkred'
                                }
                            }
                        };
                    """,
            'arpnd': """
                    function(params) {
                        if (params.data.state === 'failed') {
                            return {
                                'color': 'white',
                                'backgroundColor': 'red'
                            }
                        }
                    };
                 """,
            'vlan': """
                    function(params) {
                        if (params.data.state != 'active') {
                            return {
                                'color': 'white',
                                'backgroundColor': 'red'
                            }
                        }
                    };
                """,
            'sqPoller': """
                        function(params) {
                            if (params.data.status != 0 &&
                                params.data.status != 200) {
                                return {
                                    'color': 'white',
                                    'backgroundColor': 'red'
                                }
                            }
                        };
                        """,
            'interfaces': """
                        function(params) {
                            if (params.data.state === 'down') {
                                return {
                                    'color': 'white',
                                    'backgroundColor': 'red'
                                }
                            } else if (params.data.state === 'notConnected') {
                                return {
                                    'color': 'black',
                                    'backgroundColor': 'gray'
                                }
                            } else if (params.data.state === 'errDisabled') {
                                return {
                                    'color': 'white',
                                    'backgroundColor': 'darkred'
                                }
                            }
                        };
                        """,
            'inventory': """
                            function(params) {
                                if (params.data.status == 'absent') {
                                    return {
                                        'color': 'black',
                                        'backgroundColor': 'gray'
                                    }
                                }
                            };
                        """,
            'default': """
                        function(params) {
                                return {
                                    'backgroundColor': 'white'
                                }
                        };
                       """
        }
        style_dict['evpnVni'] = style_dict['interfaces']
        style_dict['address'] = style_dict['interfaces']

        a = style_dict.get(table, style_dict['default'])

        return JsCode(a)

    def _draw_summary_df(self, layout, query_str):
        '''Display the summary dataframe'''
        summ_df = self._run_summarize(
            namespace=self._state.namespace.split(),
            start_time=self._state.start_time,
            end_time=self._state.end_time,
            query_str=query_str)

        header_col = layout['header']
        header_col.write(
            f'<h2 style="color: darkblue; font-weight: bold;">'
            f'View: {self._state.table}</h2>',
            unsafe_allow_html=True)

        summ_ph = layout['summary']
        if not summ_df.empty:
            with summ_ph:
                st.subheader('Summary Information')
                st.dataframe(summ_df.astype(str))

    def _draw_uniq_histogram(self, layout, show_df):
        '''Display the unique histogram'''

        uniq_col = layout['uniq_col']
        if show_df.empty:
            uniq_col.info('No data to choose from')
            return

        state = self._state
        dfcols = show_df.columns.tolist()
        if (state.table == 'routes' and 'prefix' in dfcols and
                'prefixlen' not in dfcols):
            dfcols.append('prefixlen')

        dfcols = sorted((filter(lambda x: x not in ['index', 'sqvers'],
                                dfcols)))
        uniq_clicked = get_sqobject(state.table)().unique_default_column[0]
        if 'state' in dfcols:
            uniq_clicked = 'state'
        elif 'status' in dfcols:
            uniq_clicked = 'status'
        with uniq_col:
            if state.uniq_clicked not in dfcols:
                if uniq_clicked in dfcols:
                    selindex = dfcols.index(uniq_clicked)+1
                elif 'hostname' in dfcols:
                    selindex = dfcols.index('hostname') + 1
                else:
                    selindex = 1
            elif state.uniq_clicked in dfcols:
                selindex = dfcols.index(state.uniq_clicked)+1
            else:
                selindex = 1

            uniq_clicked = st.selectbox(
                'Distribution Count of', options=['-'] + dfcols,
                index=selindex, key=f'xplore_uniq_col_{state.table}',
                on_change=self._sync_uniq_col)

        scol2 = layout['uniq']

        if uniq_clicked != '-':
            uniq_df = self._run_unique(show_df,
                                       columns=uniq_clicked)
        else:
            uniq_df = pd.DataFrame()

        if not uniq_df.empty:
            if uniq_df.shape[0] > 16:
                with uniq_col:
                    st.warning(
                        f'{uniq_clicked} has cardinality of '
                        f'{uniq_df.shape[0]}, displaying top 16')
                chart = alt.Chart(
                    uniq_df.head(16),
                    title=f'{uniq_clicked} Distribution') \
                    .mark_bar(color=SUZIEQ_COLOR, tooltip=True) \
                    .encode(y=alt.Y(f'{uniq_clicked}:N',
                                    sort='-x'),
                            x='numRows')
            else:
                chart = alt.Chart(
                    uniq_df, title=f'{uniq_clicked} Distribution') \
                    .mark_bar(color=SUZIEQ_COLOR, tooltip=True) \
                    .encode(y=alt.Y(f'{uniq_clicked}:N',
                                    sort='-x'),
                            x='numRows')

            scol2.altair_chart(chart)

    def _draw_assert_df(self, layout):
        '''Display the assert dataframe'''
        if self._state.assert_clicked:
            assert_df = self._run_assert(
                start_time=self._state.start_time,
                end_time=self._state.end_time,
                namespace=self._state.namespace.split())
        else:
            assert_df = pd.DataFrame()

        if self._state.table in ['interfaces', 'ospf', 'bgp', 'evpnVni']:

            assert_expander = layout['assert'].expander(
                'Assert', expanded=self._state.assert_clicked)
            with assert_expander:
                if not assert_df.empty:
                    convert_dict = {
                        x: 'str'
                        for x in assert_df.select_dtypes('category').columns}
                    st.dataframe(data=sq_gui_style(assert_df
                                                   .astype(convert_dict),
                                                   self._state.table, True))
                elif self._state.assert_clicked:
                    st.success('Assert passed')
                else:
                    st.info('Assert not run')

    @st.cache(ttl=90)
    def _run_summarize(self, **kwargs):
        '''Get summarize dataframe for the object in question'''
        df = gui_get_df(self._state.table, self._config_file, verb='summarize',
                        **kwargs)
        # Leaving this commented to avoid future heartburn in case someone
        # tries to do this. It didn't fix the Timedelta being added to display
        # if not df.empty:
        #     if 'upTimeStat' in df.T.columns:
        #         df.T['upTimeStat'] = df.T.upTimeStat \
        #               .apply(lambda x: [str(y) for y in x])

        return df

    def _run_unique(self, df: pd.DataFrame, **kwargs) -> pd.DataFrame:
        '''Compute the distribution counts for the dataframe provided'''
        column = kwargs.pop('columns', '')
        if not df.empty:
            if column == 'prefixlen' and 'prefixlen' not in df.columns:
                # Special handling for prefixlen
                df['prefixlen'] = df['prefix'].str.split('/').str[1]

            if df.apply(lambda x: isinstance(x[column], np.ndarray),
                        axis=1).all():
                idf = df.explode(column).dropna(how='any')
            else:
                idf = df

            r = idf[column].value_counts()
            return (pd.DataFrame({column: r})
                    .reset_index()
                    .rename(columns={column: 'numRows',
                                     'index': column})
                    .sort_values(by=['numRows'], ascending=False))
        return pd.DataFrame()

    def _run_assert(self, **kwargs):
        table = self._state.table
        if table.startswith('interface'):
            df = gui_get_df(table, self._config_file, 'aver',
                            ignore_missing_peer=True,
                            result='fail', **kwargs)
        else:
            df = gui_get_df(table, self._config_file,
                            'aver', result='fail', **kwargs)
        if not df.empty:
            df.rename(columns={'assert': 'result'},
                      inplace=True, errors='ignore')
        return df



"""This module contains the base class for a PollerManager.

Classes:
    PollerManager: The duty of a pollerManager is to manage pollers data
                   In some cases it can also monitor pollers
"""
from abc import abstractmethod
from typing import Dict, List
from suzieq.poller.controller.base_controller_plugin import ControllerPlugin


class Manager(ControllerPlugin):
    """Manage and, in some cases, monitor pollers
    """

    @abstractmethod
    async def apply(self, inventory_chunks: Dict):
        """Apply the inventory chunks to the pollers

        Args:
            inventory_chunks (Dict): the portions of the global inventory
                                     to be passed to the poller
        """

    @abstractmethod
    async def launch_with_dir(self):
        """Launch a single poller writing the content of and input directory
        produced with the run-once=gather mode
        """

    @abstractmethod
    def get_n_workers(self, inventory: List[Dict]) -> int:
        """Given an inventory as input, return the required pollers to query
        all the devices in it.

        Args:
            inventory (List[Dict]): the inventory to be splitted across the
                                    poller instances

        Returns:
            int: number of desired workers
        """
    @classmethod
    def default_type(cls) -> str:
        return 'static'

    @classmethod
    def get_data_model(cls):
        """This is only temporary. In future release I will add manager
        validation via pydantic
        """
        raise NotImplementedError



from typing import List
from suzieq.sqobjects.basicobj import SqObject


def get_tables() -> List[str]:
    """Return a list of all supported tables/services

    Returns:
        List[str]: sorted list of all supported tables/services
    """
    tables = SqObject.get_plugins()
    return sorted(tables.keys())


def get_sqobject(table_name: str):
    '''Return the object to operate on the table requested

    :param table_name: str, name of table for which object is requested
    :returns:
    :rtype: The child class of SqObject that corresponds to the table
            or raises ModuleNotFoundError if the table is not found

    '''

    tables = SqObject.get_plugins()
    if table_name not in tables:
        if table_name in ['interface', 'route', 'mac', 'table']:
            # Handle singular/plural conversion
            table_name += 's'

    if table_name not in tables:
        raise ModuleNotFoundError(f'Table {table_name} not found')

    return tables[table_name]


__all__ = get_tables() + ['get_sqobject']



import pandas as pd

from suzieq.engines.pandas.engineobj import SqPandasEngine


class NamespaceObj(SqPandasEngine):
    '''Backend class to handle ops on virtual table, namespace, with pandas'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'namespace'

    def get(self, **kwargs):
        """Get the information requested"""

        kwargs.get('view', self.iobj.view)
        columns = kwargs.pop('columns', ['default'])
        kwargs.pop('addnl_fields', [])
        user_query = kwargs.pop('query_str', '')
        os = kwargs.pop('os', [])
        model = kwargs.pop('model', [])
        vendor = kwargs.pop('vendor', [])
        os_version = kwargs.pop('version', [])
        namespace = kwargs.pop('namespace', [])

        drop_cols = []
        fields = self.schema.get_display_fields(columns)

        if os or model or vendor or os_version:
            devdf = self._get_table_sqobj('device').get(
                columns=['namespace', 'hostname', 'os', 'model', 'vendor',
                         'version'],
                os=os, model=model, vendor=vendor, version=os_version,
                namespace=namespace, **kwargs)
        else:
            devdf = self._get_table_sqobj('device').get(
                columns=['namespace', 'hostname'], namespace=namespace,
                **kwargs)

        if devdf.empty:
            return pd.DataFrame()

        namespace = devdf.namespace.unique().tolist()
        dev_nsgrp = devdf.groupby(['namespace'])

        # Get list of namespaces we're polling
        pollerdf = self._get_table_sqobj('sqPoller') \
            .get(columns=['namespace', 'hostname', 'service', 'status',
                          'timestamp'],
                 namespace=namespace)

        if pollerdf.empty:
            return pd.DataFrame()

        pollerdf = devdf.merge(pollerdf, on=['namespace', 'hostname'])
        nsgrp = pollerdf.groupby(by=['namespace'])
        pollerns = sorted(pollerdf.namespace.unique().tolist())
        newdf = pd.DataFrame({
            'namespace': pollerns,
            'deviceCnt': dev_nsgrp['hostname'].nunique().tolist(),
            'serviceCnt': nsgrp['service'].nunique().tolist()
        })
        errsvc_df = pollerdf.query('status != 0 and status != 200') \
                            .groupby(by=['namespace'])['service'] \
                            .nunique().reset_index()
        newdf = newdf.merge(errsvc_df, on=['namespace'], how='left',
                            suffixes=['', '_y']) \
            .rename({'service': 'errSvcCnt'}, axis=1) \
            .fillna({'errSvcCnt': 0})
        newdf['errSvcCnt'] = newdf['errSvcCnt'].astype(int)

        # What protocols exist
        for table, fld in [('ospf', 'hasOspf'), ('bgp', 'hasBgp'),
                           ('evpnVni', 'hasVxlan'), ('mlag', 'hasMlag')]:
            df = self._get_table_sqobj(table) \
                .get(namespace=pollerns, columns=['namespace', 'hostname'])
            if df.empty:
                newdf[fld] = False
                continue

            df = df.merge(devdf, on=['namespace', 'hostname'])
            if df.empty:
                newdf[fld] = False
                continue

            gotns = df.namespace.unique().tolist()
            newdf[fld] = newdf.apply(
                lambda x, y: x.namespace in y,
                axis=1, args=(gotns,))

        if 'sqvers' in fields:
            newdf['sqvers'] = self.schema.version
        newdf['active'] = True
        newdf = self._handle_user_query_str(newdf, user_query)

        # Look for the rest of info only in selected namepaces
        newdf['lastUpdate'] = nsgrp['timestamp'].max() \
            .reset_index()['timestamp']

        newdf = self._handle_user_query_str(newdf, user_query)
        return newdf.drop(columns=drop_cols)[fields]

    def summarize(self, **kwargs):
        '''Summarize for network

        Summarize for network is a bit different from the rest of the
        summaries because its not grouped by namespace.
        '''

        df = self.get(**kwargs)

        if df.empty:
            return df

        self.ns = pd.DataFrame({
            'namespacesCnt': [df.namespace.count()],
            'servicePerNsStat': [[df.serviceCnt.min(), df.serviceCnt.max(),
                                  df.serviceCnt.median()]],
            'nsWithMlagCnt': [df.loc[df.hasMlag].shape[0]],
            'nsWithBgpCnt': [df.loc[df.hasBgp].shape[0]],
            'nsWithOspfCnt': [df.loc[df.hasOspf].shape[0]],
            'nsWithVxlanCnt': [df.loc[df.hasVxlan].shape[0]],
            'nsWithErrsvcCnt': [df.loc[df.errSvcCnt != 0].shape[0]],
        })

        # At this point we have a single row with index 0, so rename
        return self.ns.T.rename(columns={0: 'summary'})



from dataclasses import dataclass
from typing import Tuple
from urllib.parse import quote

import pandas as pd
import streamlit as st
import graphviz
from st_aggrid import (GridOptionsBuilder, AgGrid, GridUpdateMode,
                       JsCode)

from suzieq.sqobjects import get_sqobject
from suzieq.gui.stlit.guiutils import (gui_get_df, set_def_aggrid_options,
                                       display_help_icon, get_base_url,
                                       get_session_id, SuzieqMainPages)
from suzieq.gui.stlit.pagecls import SqGuiPage


@dataclass
class PathSessionState:
    '''Session state for status page'''
    page: str = SuzieqMainPages.PATH.value
    namespace: str = ''
    source: str = ''
    dest: str = ''
    start_time: str = ''
    end_time: str = ''
    show_ifnames: bool = False
    vrf: str = ''


class PathPage(SqGuiPage):
    '''Page for Path trace page'''
    _title: str = SuzieqMainPages.PATH.value
    _state: PathSessionState = PathSessionState()
    _failed_dfs = {}
    _pathobj = None
    _path_df: pd.DataFrame = None
    _config_file: str = st.session_state.get('config_file', '')

    @property
    def add_to_menu(self) -> bool:
        return True

    def build(self):
        self._get_state_from_url()
        self._create_sidebar()
        layout = self._create_layout()
        self._render(layout)

    def _create_layout(self) -> dict:

        pgbar = st.empty()
        summary = st.container()
        summcol, _, pathcol = summary.columns([3, 1, 10])
        with summary:
            with summcol:
                summ_ph = st.empty()
                legend_ph = st.container()
            with pathcol:
                fw_ph = st.empty()
        table_expander = st.empty()
        failed_tables = st.empty()
        return ({
            'pgbar': pgbar,
            'summary': summ_ph,
            'legend': legend_ph,
            'fw_path': fw_ph,
            'table': table_expander,
            'failed_tables': failed_tables,
        })

    def _create_sidebar(self) -> None:

        state = self._state
        devdf = gui_get_df('device', config_file=self._config_file,
                           columns=['namespace'])
        if devdf.empty:
            st.error('Unable to retrieve any namespace info')
            st.stop()

        namespaces = [''] + sorted(devdf.namespace.unique().tolist())
        if self._state.namespace:
            nsidx = namespaces.index(self._state.namespace)
        else:
            nsidx = 0

        url = '&amp;'.join([
            f'{get_base_url()}?page=Help&session={get_session_id()}',
            'help=yes',
            'help_on=Path',
        ])
        display_help_icon(url)
        with st.sidebar:
            with st.form(key='trace'):
                state.namespace = st.selectbox(
                    'Namespace', namespaces, key='path_namespace',
                    index=nsidx)
                state.source = st.text_input('Source IP',
                                             key='path_source',
                                             value=state.source)
                state.dest = st.text_input('Dest IP', value=state.dest,
                                           key='path_dest')

                state.vrf = st.text_input('VRF', value=state.vrf,
                                          key='path_vrf')
                state.start_time = st.text_input('Start Time',
                                                 value=state.start_time,
                                                 key='path_start_time')
                state.end_time = st.text_input('End Time',
                                               value=state.end_time,
                                               key='path_end_time')

                _ = st.form_submit_button('Trace', on_click=self._sync_state)

            state.show_ifnames = st.checkbox('Show in/out interface names',
                                             value=state.show_ifnames,
                                             key='path_show_ifnames',
                                             on_change=self._sync_state)
            _ = st.button(
                'Source <-> Dest', key='path_swap',
                on_click=self._swap_src_dest)

    def _swap_src_dest(self):
        state = self._state
        state.source, state.dest = state.dest, state.source
        self._sync_state()

    def _sync_state(self) -> None:
        wsstate = st.session_state
        state = self._state

        state.namespace = wsstate.path_namespace
        state.vrf = wsstate.path_vrf
        state.start_time = wsstate.path_start_time
        state.end_time = wsstate.path_end_time
        state.show_ifnames = wsstate.path_show_ifnames
        self._save_page_url()

    def _render(self, layout: dict) -> None:
        '''Compute the path, render all objects'''

        state = self._state
        missing = [x for x in [state.source,
                               state.dest, state.namespace] if not x]
        if missing:
            st.warning('Set namespace, source and destination in the sidebar '
                       'to do a path trace ')
            st.stop()

        layout['pgbar'].progress(0)
        self._pathobj = get_sqobject(
            'path')(config_file=self._config_file,
                    start_time=state.start_time,
                    end_time=state.end_time)
        try:
            df, summ_df = self._get_path(forward_dir=True)
            rdf = getattr(self._pathobj.engine, '_rdf', pd.DataFrame())
            if not rdf.empty:
                # Something did get computed
                self._path_df = df
        except Exception as e:  # pylint: disable=broad-except
            st.error(f'Invalid Input: {str(e)}')
            layout['pgbar'].progress(100)
            self._path_df = pd.DataFrame()
            st.stop()
            return

        layout['pgbar'].progress(40)

        if df.empty:
            layout['pgbar'].progress(100)
            st.info(f'No path to trace between {self._state.source} and '
                    f'{self._state.dest}')
            st.stop()
            return

        self._get_failed_data(state.namespace, layout['pgbar'])

        g = self._build_graphviz_obj(state.show_ifnames, df)
        layout['pgbar'].progress(100)
        # if not rev_df.empty:
        #     rev_g = build_graphviz_obj(state, rev_df)

        layout['summary'].dataframe(data=summ_df.astype(str))
        with layout['legend']:
            st.info('''Color Legend''')
            st.markdown('''
<b style="color:blue">Blue Lines</b> => L2 Hop(non-tunneled)<br>
<b>Black Lines</b> => L3 Hop<br>
<b style="color:purple">Purple lines</b> => Tunneled Hop<br>
<b style="color:red">Red Lines</b> => Hops with Error<br>
''', unsafe_allow_html=True)

        layout['fw_path'].graphviz_chart(g, use_container_width=True)
        # rev_ph.graphviz_chart(rev_g, use_container_width=True)

        with layout['table']:
            table_expander = st.expander('Path Table', expanded=True)
            with table_expander:
                self._draw_aggrid_df(df)

    def _draw_aggrid_df(self, df):

        gb = GridOptionsBuilder.from_dataframe(df)
        gb.configure_pagination(paginationPageSize=25)

        gb.configure_default_column(floatingFilter=True)
        if 'error' in df.columns:
            jscode = self._aggrid_style()
            gb.configure_column("error", floatingFilter=True,
                                cellStyle=jscode, editable=False)

        gb.configure_grid_options(
            domLayout='normal', preventDefaultOnContextMenu=True)

        gridOptions = gb.build()
        gridOptions = set_def_aggrid_options(gridOptions)
        gridOptions['getRowStyle'] = self._aggrid_style_rows(df)

        _ = AgGrid(
            df,
            gridOptions=gridOptions,
            allow_unsafe_jscode=True,
            update_mode=GridUpdateMode.NO_UPDATE,
            theme='streamlit',
        )

    def _aggrid_style_rows(self, df: pd.DataFrame):
        '''Style the cells based on value'''

        if 'error' in df.columns:
            a = """
                function(params) {
                    if (params.data.error !== '') {
                        return {
                            'color': 'white',
                            'backgroundColor': 'red'
                        }
                    }
                };
                """
        else:
            a = """
            function(params) {
                    return {
                        'backgroundColor': 'white'
                    }
            };
            """
        return JsCode(a)

    def _get_path(self, forward_dir: bool = True) -> Tuple[pd.DataFrame,
                                                           pd.DataFrame]:
        '''Get path & summary df'''
        if forward_dir:
            df = self._pathobj.get(namespace=[self._state.namespace],
                                   src=self._state.source,
                                   dest=self._state.dest,
                                   vrf=self._state.vrf)

            summ_df = self._path_summarize(df)
        return df, summ_df

    def _get_failed_data(self, namespace: str, pgbar) -> None:
        '''Get interface/mlag/routing protocol states that are failed'''

        progress = 40
        for i, entry in enumerate([
                {'name': 'device', 'query': 'status == "dead"'},
                {'name': 'interfaces',
                 'query': 'state == "down" and adminState == "up"'},
                {'name': 'mlag',
                 'query': '(state != "active") or (mlagSinglePortsCnt != 0 or '
                 'mlagErrorPortsCnt != 0)'},
                {'name': 'ospf', 'query': 'adjState == "other"'},
                {'name': 'bgp', 'query': 'state == "NotEstd"'}
        ]):
            df = gui_get_df(
                entry['name'], self._config_file, namespace=[namespace])
            if not df.empty and (entry.get('query', '')):
                df = df.query(entry['query']).reset_index(drop=True)
                self._failed_dfs[entry['name']] = df
            pgbar.progress(progress + i*10)

    def _path_summarize(self, path_df: pd.DataFrame) -> pd.DataFrame:
        '''Summarizes the path, a copy of the function in pandas/path.py

        Until we support a mechanism to avoid running path twice to get
        a summary, this is the best we can do to avoid running path twice
        '''

        if path_df.empty:
            return pd.DataFrame()

        namespace = path_df.namespace.iloc[0]
        ns = {}
        ns[namespace] = {}

        perhopEcmp = path_df.query('hopCount != 0') \
                            .groupby(by=['hopCount'])['hostname']
        ns[namespace]['totalPaths'] = path_df['pathid'].max()
        ns[namespace]['perHopEcmp'] = perhopEcmp.nunique().tolist()
        ns[namespace]['maxPathLength'] = path_df.groupby(by=['pathid'])[
            'hopCount'].max().max()
        ns[namespace]['avgPathLength'] = path_df.groupby(by=['pathid'])[
            'hopCount'].max().mean()
        ns[namespace]['uniqueDevices'] = path_df['hostname'].nunique()
        ns[namespace]['mtuMismatch'] = not all(path_df['mtuMatch'])
        ns[namespace]['usesOverlay'] = any(path_df['overlay'])
        ns[namespace]['pathMtu'] = min(
            path_df.query('iif != "lo"')['inMtu'].min(),
            path_df.query('iif != "lo"')['outMtu'].min())

        summary_fields = ['totalPaths', 'perHopEcmp', 'maxPathLength',
                          'avgPathLength', 'uniqueDevices', 'pathMtu',
                          'usesOverlay', 'mtuMismatch']
        return pd.DataFrame(ns).reindex(summary_fields, axis=0) \
                               .convert_dtypes()

    # pylint: disable=too-many-statements
    @st.cache(max_entries=10, allow_output_mutation=True)
    def _build_graphviz_obj(self, show_ifnames: bool, df: pd.DataFrame):
        '''Return a graphviz object'''

        graph_attr = {'splines': 'polyline', 'layout': 'dot'}
        if show_ifnames:
            graph_attr.update({'nodesep': '1.0'})

        g = graphviz.Digraph(graph_attr=graph_attr,
                             name='Hover over arrow head for edge info')

        hostset = set()
        for hostgroup in df.groupby(by=['hopCount']) \
                           .hostname.unique().tolist():
            with g.subgraph() as s:
                s.attr(rank='same')
                for hostname in hostgroup:
                    if hostname in hostset:
                        continue
                    hostset.add(hostname)
                    debugURL = '&amp;'.join([
                        f'{get_base_url()}?page={quote("Path-Debug")}',
                        'lookupType=hop',
                        f'namespace={quote(df.namespace[0])}',
                        f'session={quote(get_session_id())}',
                        f'hostname={quote(hostname)}',
                    ])
                    tooltip, color = self._get_node_tooltip_color(
                        hostname)
                    s.node(hostname, tooltip=tooltip, color=color,
                           URL=debugURL, target='_graphviz', shape='box')

        pathid = 0
        prevrow = None
        connected_set = set()

        df['nextPathid'] = df.pathid.shift(-1).fillna('0').astype(int)
        for row in df.itertuples():
            if row.pathid != pathid:
                prevrow = row
                pathid = row.pathid
                continue
            conn = (prevrow.hostname, row.hostname)
            if conn not in connected_set:
                if row.overlay:
                    path_type = 'underlay'
                    color = 'purple'
                elif prevrow.isL2:
                    path_type = 'l2'
                    color = 'blue'
                else:
                    path_type = 'l3'
                    color = 'black'

                if not row.mtuMatch:
                    color = 'red'
                    error = 'MTU mismatch'
                    err_pfx = ', '
                else:
                    error = ''
                    err_pfx = ''

                tdf = pd.DataFrame({
                    'pathType': path_type,
                    'protocol': [prevrow.protocol],
                    'ipLookup': [prevrow.ipLookup],
                    'vtepLookup': [prevrow.vtepLookup],
                    'macLookup': [prevrow.macLookup],
                    'nexthopIp': [prevrow.nexthopIp],
                    'vrf': [prevrow.vrf],
                    'mtu': [f'{prevrow.outMtu} -> {row.inMtu}'],
                    'oif': [prevrow.oif],
                    'iif': [row.iif]})
                rowerr = getattr(prevrow, 'error', '')
                if rowerr:
                    error += f"{err_pfx}{rowerr}"
                if error:
                    err_pfx = ', '
                if row.nextPathid != row.pathid:
                    # We need to capture any errors on the dest node as well
                    destnode_error = getattr(row, 'error', '')
                    if destnode_error:
                        error += f'{err_pfx}{destnode_error}'

                if error:
                    tdf['error'] = error
                    color = 'red'
                tooltip = '\n'.join(tdf.T.to_string(
                    justify='right').split('\n')[1:])
                debugURL = '&amp;'.join([
                    f'{get_base_url()}?page={quote("Path-Debug")}',
                    'lookupType=edge',
                    f'namespace={quote(row.namespace)}',
                    f'session={quote(get_session_id())}',
                    f'hostname={quote(prevrow.hostname)}',
                    f'vrf={quote(prevrow.vrf)}',
                    f'vtepLookup-{prevrow.vtepLookup}',
                    f'ifhost={quote(row.hostname)}',
                    f'ipLookup={quote(prevrow.ipLookup)}',
                    f'oif={quote(prevrow.oif)}',
                    f'macaddr={quote(prevrow.macLookup or "")}',
                    f'nhip={quote(prevrow.nexthopIp)}',
                ])
                if show_ifnames:
                    g.edge(prevrow.hostname, row.hostname, color=color,
                           label=str(row.hopCount), URL=debugURL,
                           edgetarget='_graphviz', tooltip=tooltip,
                           taillabel=prevrow.oif, headlabel=row.iif,
                           penwidth='2.0',
                           )
                else:
                    g.edge(prevrow.hostname, row.hostname, color=color,
                           label=str(row.hopCount), URL=debugURL,
                           edgetarget='_graphviz', penwidth='2.0',
                           tooltip=tooltip
                           )

                connected_set.add(conn)
            prevrow = row
        df.drop(columns=['nextPathid'], inplace=True, errors='ignore')
        return g

    def _get_node_tooltip_color(self, hostname: str) -> str:
        '''Get tooltip and node color for node based on various tables'''

        ttip = {'title': ['Failed entry count']}
        for tbl, df in self._failed_dfs.items():
            ttip.update({tbl: [df.loc[df.hostname == hostname]
                               .hostname.count()]})
        tdf = pd.DataFrame(ttip)
        if tdf.select_dtypes(include='int64').max().max() > 0:
            color = 'red'
        else:
            color = 'black'
        return('\n'.join(tdf.T.to_string().split('\n')[1:]), color)

    def _aggrid_style(self):
        '''Style the cells based on value'''
        a = """
            function(params) {
                if (params.error != '') {
                    return {
                        'color': 'white',
                        'backgroundColor': 'darkred'
                    }
                }
            };
            """
        return JsCode(a)

    @staticmethod
    def _highlight_erroneous_rows(row):
        '''Highlight rows with error in them'''
        if getattr(row, 'error', '') != '':
            return ["background-color: red; color: white"]*len(row)
        else:
            return [""]*len(row)



import pandas as pd
from suzieq.sqobjects.basicobj import SqObject


class DevconfigObj(SqObject):
    '''The object providing access to the device config table'''

    def __init__(self, **kwargs):
        super().__init__(table='devconfig', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns',
                                'query_str']

    def unique(self, **kwargs) -> pd.DataFrame:
        return pd.DataFrame(
            {'error': ['Unique not supported for Device Config']})



from suzieq.sqobjects.basicobj import SqObject


class TopologyObj(SqObject):
    '''The object providing access to the virtual table: topology'''

    def __init__(self, **kwargs):
        '''Init routine'''
        super().__init__(table='topology', **kwargs)
        self._sort_fields = ["namespace", "hostname", "ifname"]
        self._cat_fields = []
        self._valid_get_args = ['namespace', 'hostname', 'columns',
                                'polled', 'ifname', 'via', 'vrf', 'asn',
                                'area', 'peerHostname', 'afiSafi', 'query_str']
        self._valid_summarize_args = ['namespace', 'hostname', 'via', 'vrf',
                                      'asn', 'area', 'query_str']
        self._valid_arg_vals = {
            'polled': ['True', 'False', 'true', 'false', '']
        }
        self._valid_summarize_args = ['namespace', 'hostname', 'via',
                                      'query_str']



import re
from datetime import datetime
from collections import defaultdict
from json import loads
from dateparser import parse
import numpy as np

from suzieq.poller.worker.services.service import Service
from suzieq.shared.utils import (get_timestamp_from_junos_time,
                                 expand_ios_ifname, expand_nxos_ifname)
from suzieq.shared.utils import convert_macaddr_format_to_colon
from suzieq.shared.utils import MISSING_SPEED, NO_SPEED, MISSING_SPEED_IF_TYPES


# pylint: disable=too-many-statements

class InterfaceService(Service):
    """Service class for interfaces. Cleanup of data is specific"""

    def __init__(self, name, defn, period, stype, keys, ignore_fields, schema,
                 queue, db_access, run_once="forever"):

        super().__init__(name, defn, period, stype, keys, ignore_fields,
                         schema, queue, db_access, run_once)
        # Thanks to JunOS obsolete timestamp, we cannot use the
        # statusChangeTimestamp field directly. We use an artificial field
        # called statusChangeTimestamp1 to ensure we capture changes
        self.ignore_fields.append("statusChangeTimestamp")

    def _assign_vrf(self, entry, entry_dict):
        '''Uses the interface List to assign interfaces to VRFs'''
        if entry['type'] == 'vrf':
            if entry['ifname'] != 'default':
                for intf in entry['_interfaceList']:
                    if intf in entry_dict:
                        entry_dict[intf]['master'] = entry['ifname']
                    else:
                        entry_dict[intf] = {'vrf': entry['ifname']}

    def _get_missing_speed_value(self, entry):
        '''
        Return the correct value for an interface without a valid speed
        '''
        if entry.get('type', '') not in MISSING_SPEED_IF_TYPES:
            return NO_SPEED
        return MISSING_SPEED

    def _speed_field_check(self, entry, missing_speed_indicator):
        """
        Return a missing-speed value if the speed is invalid,
        the interface speed otherwise

        MUST be called after the type's been fixed correctly
        """
        if entry.get('speed', missing_speed_indicator) \
           == missing_speed_indicator:
            return self._get_missing_speed_value(entry)
        return entry['speed']

    def _common_speed_field_value(self, entry):
        """Return speed value or a missing value for common retrieved data"""
        return self._speed_field_check(entry, 0)

    def _textfsm_valid_speed_value(self, entry):
        """Return speed value or a missing value for textfsm retrieved data"""
        return self._speed_field_check(entry, '')

    def _clean_eos_data(self, processed_data, _):
        """Clean up EOS interfaces output"""

        entry_dict = defaultdict(dict)
        drop_indices = []
        vlan_entries = {}       # Needed to fix the anycast MAC entries

        # pylint: disable=too-many-nested-blocks
        for i, entry in enumerate(processed_data):

            if entry['type'] == 'vrf':
                if entry['ifname'] == 'default':
                    drop_indices.append(i)
                    continue

                self._assign_vrf(entry, entry_dict)
                entry['macaddr'] = '00:00:00:00:00:00'
                entry['master'] = ''
                entry['state'] = 'up'
                entry['adminState'] = 'up'
                entry['mtu'] = 1500
                entry['speed'] = NO_SPEED
                continue

            if entry['type'] == 'varp':
                for elem, val in vlan_entries.items():
                    ventry = val
                    amac = ventry.get('_anycastMac', '')
                    if amac:
                        ventry['interfaceMac'] = ventry['macaddr']
                        ventry['macaddr'] = amac
                    else:
                        # Found this format in some of the real life deployment
                        amacs = entry.get('_virtualMacs', []) or []
                        for amac in amacs:
                            if amac.get('macType', '') == 'varp':
                                ventry['interfaceMac'] = ventry['macaddr']
                                ventry['macaddr'] = amac['macAddress']
                                break
                drop_indices.append(i)
                continue

            if not entry_dict[entry['ifname']]:
                entry_dict[entry['ifname']] = entry
            elif 'vrf' in entry_dict[entry['ifname']]:
                entry['master'] = entry_dict[entry['ifname']]['vrf']
                entry_dict[entry['ifname']] = entry

            ts = entry["statusChangeTimestamp"]
            if ts:
                entry["statusChangeTimestamp"] = int(float(ts) * 1000)
            else:
                entry["statusChangeTimestamp"] = 0
            # artificial field for comparison with previous poll result
            entry["statusChangeTimestamp1"] = entry.get(
                "statusChangeTimestamp", '')
            if entry["type"] == "portChannel":
                entry["type"] = "bond"
            words = entry.get("master", "")
            if words:
                words = words.split()
                if words[-1].strip().startswith("Port-Channel"):
                    entry["type"] = "bond_slave"
                    entry["master"] = words[-1].strip()
            entry["lacpBypass"] = (entry["lacpBypass"] is True)
            if entry["forwardingModel"] == "bridged":
                entry["master"] = "bridge"  # Convert it for Linux model
                del entry["forwardingModel"]

            if entry['type']:
                entry['type'] = entry['type'].lower()

            # Vlan is gathered as a list for VXLAN interfaces. Fix that
            if entry["type"] == "vxlan":
                # We capture the glory in evpnVni of this interface
                entry["vlan"] = 0
                entry['mtu'] = -1
                entry['macaddr'] = '00:00:00:00:00:00'

            if entry['type'] == 'vlan' and entry['ifname'].startswith('Vlan'):
                entry['vlan'] = int(entry['ifname'].split('Vlan')[1])
                vlan_entries[entry['ifname']] = entry

            adm_state = entry.get('adminState', 'down')
            if adm_state == 'notconnect':
                entry['reason'] = 'notconnect'
                entry['adminState'] = 'down'
                entry['state'] = 'notConnected'
            elif adm_state == 'errdisabled':
                entry['reason'] = 'errdisabled'
                entry['adminState'] = 'down'
                entry['state'] = 'errDisabled'
            elif adm_state == 'connected':
                entry['adminState'] = 'up'

            speed = self._common_speed_field_value(entry)
            if speed != MISSING_SPEED:
                speed = int(speed / 1000000)
            entry['speed'] = speed

            tmpent = entry.get("ipAddressList", [[]])
            if not tmpent:
                continue

            if entry['type'] == "loopback":
                entry['macaddr'] = '00:00:00:00:00:00'

            munge_entry = tmpent[0]
            if munge_entry:
                new_list = []
                primary_ip = (
                    munge_entry["primaryIp"]["address"]
                    + "/"
                    + str(munge_entry["primaryIp"]["maskLen"])
                )
                new_list.append(primary_ip)
                for elem in munge_entry["secondaryIpsOrderedList"]:
                    ip = elem["address"] + "/" + str(elem["maskLen"])
                    new_list.append(ip)
                if 'virtualIp' in munge_entry:
                    elem = munge_entry['virtualIp']
                    if elem["address"] != "0.0.0.0":
                        ip = f'{elem["address"]}/{str(elem["maskLen"])}'
                        new_list.append(ip)
                entry["ipAddressList"] = new_list

            # ip6AddressList is formatted as a dict, not a list by EOS
            munge_entry = entry.get("ip6AddressList", [{}])
            if munge_entry:
                new_list = []
                for elem in munge_entry.get("globalUnicastIp6s", []):
                    new_list.append(elem["subnet"])
                entry["ip6AddressList"] = new_list

        if drop_indices:
            processed_data = np.delete(processed_data, drop_indices).tolist()

        return processed_data

    def _clean_cumulus_data(self, processed_data, _):
        """We have to merge the appropriate outputs of two separate commands"""

        for entry in processed_data:
            if entry.get('hardware', '') == 'ether':
                entry['type'] = 'ethernet'

            if entry['adminState'] == "down":
                entry['state'] = "down"

            if entry['type'] == 'ether':
                entry['type'] = 'ethernet'
            if entry.get('type', '') == 'vxlan':
                entry['speed'] = NO_SPEED

            if not entry['linkUpCnt']:
                entry['linkUpCnt'] = 0
            if not entry['linkDownCnt']:
                entry['linkDownCnt'] = 0

            entry["numChanges"] = (int(entry["linkUpCnt"]) +
                                   int(entry["linkDownCnt"]))
            entry['state'] = entry['state'].lower()
            if entry["state"] == "up":
                ts = entry["linkUpTimestamp"]
            else:
                ts = entry["linkDownTimestamp"]
            if "never" in ts or not ts:
                ts = 0
            else:
                ts = int(
                    datetime.strptime(
                        ts.strip(), "%Y/%m/%d %H:%M:%S.%f"
                    ).timestamp()
                    * 1000
                )
            entry["statusChangeTimestamp"] = ts
            # artificial field for comparison with previous poll result
            entry["statusChangeTimestamp1"] = entry.get(
                "statusChangeTimestamp", '')

            if '(' in entry['master']:
                entry['master'] = entry['master'].replace(
                    '(', '').replace(')', '')

            # Lowercase the master value thanks to SoNIC
            entry['master'] = entry.get('master', '').lower()
            if entry['ip6AddressList'] and 'ip6AddressList-_2nd' in entry:
                # This is because textfsm adds peer LLA as well
                entry['ip6AddressList'] = entry['ip6AddressList-_2nd']

            if 'type-_2nd' in entry:
                entry['type'] = entry['type-_2nd']

            del entry["linkUpCnt"]
            del entry["linkDownCnt"]
            del entry["linkUpTimestamp"]
            del entry["linkDownTimestamp"]
            del entry["vrf"]

            entry['speed'] = self._textfsm_valid_speed_value(entry)

        return processed_data

    def _clean_junos_data(self, processed_data, raw_data):
        """Cleanup IP addresses and such"""

        def fix_junos_speed(entry):
            speed = str(self._common_speed_field_value(entry))
            if speed.endswith('mbps'):
                speed = int(speed.split('mb')[0])
            elif speed.endswith('Gbps'):
                speed = int(speed.split('Gb')[0])*1000
            elif speed == 'Unlimited':
                speed = self._get_missing_speed_value(entry)
            return speed

        entry_dict = defaultdict(dict)
        drop_indices = []
        new_entries = []

        for i, entry in enumerate(processed_data):

            ifname = entry.get('ifname', '')
            if not entry.get('macaddr', ''):
                entry['macaddr'] = '00:00:00:00:00:00'

            entry['type'] = entry.get('type', '').lower()

            if entry['type'] in ['vrf', 'virtual-router']:
                entry['type'] = 'vrf'
                self._assign_vrf(entry, entry_dict)
                entry['state'] = entry['adminState'] = 'up'
                entry['mtu'] = 1500
                if ifname == 'default':
                    drop_indices.append(i)
                entry['type'] = 'vrf'  # VMX uses virtual-router for VRF
                entry['speed'] = NO_SPEED
                continue

            if entry.get('description', '') == 'None':
                entry['description'] = ''

            if ifname.startswith(('gre', 'ipip', 'lc-', 'lsi', 'pc-', 'bme',
                                  'cbp', 'pimd', 'pime', 'pip', 'pd',
                                  'pd', 'fxp', 'mtun', 'tap', 'bcm', 'jsrv',
                                  'mo-', 'ixgbe')):
                entry['type'] = 'internal'

            if ifname.startswith('irb.'):
                entry['type'] = 'vlan'
            elif ifname == 'irb':
                entry['type'] = 'internal'
            elif ifname == 'dsc':
                entry['type'] = 'null'

            if not entry['type']:
                entry['type'] = 'internal'
                entry['mtu'] = 65536
                entry['speed'] = fix_junos_speed(entry)
                continue

            if entry['type'] == 'virtual-switch':
                # TODO: Handle this properly
                continue

            if entry['type'] == 'vxlan-tunnel-endpoint':
                entry['type'] = 'vtep'

            if entry.get('_minLinksBond', None) is not None:
                entry['type'] = 'bond'

            if entry.get('type', '') == 'vpls':
                entry['state'] = 'up'
                entry['adminState'] = 'up'

            if not entry_dict[ifname]:
                entry_dict[ifname] = entry

            if entry.get('mtu', 0) == 'Unlimited':
                entry['mtu'] = 65536

            entry['mtu'] = int(entry.get('mtu', 0))

            if 'statusChangeTimestamp' in entry:
                ts = get_timestamp_from_junos_time(
                    entry['statusChangeTimestamp'],
                    raw_data[0]['timestamp']/1000)
                entry['statusChangeTimestamp'] = int(ts)

            if entry.get('master', '') == 'Ethernet-Bridge':
                entry['master'] = 'bridge'
            else:
                entry['master'] = ''

            if entry.get('type', '') == 'ethernet-bridge':
                # This is an MX device
                entry['type'] = 'ethernet'
                entry['master'] = 'bridge'

                # extract VLAN
                vlan_tag = entry.get('_vlanTag', [])
                if vlan_tag:
                    words = vlan_tag[0].split('0x8100.')
                    vlan = words[1].split(')')[0]
                else:
                    vlan = 0

                entry['vlan'] = int(vlan)
            entry['speed'] = fix_junos_speed(entry)

            # Process the logical interfaces which are too deep to be parsed
            # efficiently by the parser right now
            for lentry in entry.get('_logIf', []) or []:
                lifname = lentry.get('name', [{}])[0].get('data', '')
                if not lifname:
                    continue
                if '.' in lifname:
                    _, vlan = lifname.split('.')
                    # This order is important, don't change; irb is an internal
                    # interface
                    if ifname == "irb":
                        iftype = 'vlan'
                    elif entry['type'] == 'internal':
                        iftype = 'internal'
                    else:
                        iftype = 'subinterface'
                    vlan = int(vlan)
                    if vlan > 4095:
                        vlan = 0
                else:
                    vlan = 0
                    iftype = entry['type']

                if iftype == 'subinterface':
                    speed = lentry \
                        .get('logical-interface-bandwidth', [{}])[0] \
                        .get('data', entry.get('speed', MISSING_SPEED))
                else:
                    speed = lentry \
                        .get('logical-interface-bandwidth', [{}])[0] \
                        .get('data', NO_SPEED)

                if not entry_dict[lifname]:
                    entry_dict[lifname] = {
                        'ifname': lifname,
                        'type': iftype,
                        'vlan': vlan,
                        'speed': speed,
                        'master': ifname,
                        'state': 'up',
                        'adminState': 'up',
                        'mtu': entry.get('mtu', 0),
                        'statusChangeTimestamp':
                        entry.get('statusChangeTimestamp', 0),
                    }

                afis = lentry.get('address-family', []) or []
                no_inet = True
                v4addresses = []
                v6addresses = []
                macaddr = lentry.get('logical-interface-vgw-v4-mac', [{}])[0] \
                    .get('data', '')
                if not macaddr:
                    macaddr = lentry.get('logical-interface-mac', [{}])[0] \
                        .get('data', '')
                if not macaddr:
                    macaddr = entry.get('macaddr', '00:00:00:00:00:00')

                addrlist = []
                for elem in afis:
                    afi_mtu = elem.get('mtu', [{}])[0].get(
                        'data', entry['mtu'])
                    afi = elem.get('address-family-name', [{}])[0] \
                        .get('data', '')
                    if afi == 'inet':
                        no_inet = False
                        addrlist += elem.get('interface-address', [])
                    elif afi == 'inet6':
                        no_inet = False
                        addrlist += elem.get('interface-address', [])
                    if afi == "aenet":
                        master = elem.get("ae-bundle-name", [{}])[0] \
                            .get("data", "")
                        entry['master'] = master.split('.')[0]
                        entry['type'] = 'bond_slave'
                    if afi in ["eth-switch", "bridge"]:
                        addrlist = []
                        entry['master'] = 'bridge'
                        continue

                if no_inet:
                    continue

                vrf = entry_dict.get(ifname, {}).get('vrf', '')

                new_entry = {'ifname': lifname,
                             'mtu': afi_mtu,
                             'type': iftype,
                             'speed': speed,
                             'vlan': vlan,
                             'master': vrf,
                             'macaddr': macaddr,
                             'adminState': 'up',
                             'description': entry['description'],
                             'state': 'up',
                             'statusChangeTimestamp':
                             entry['statusChangeTimestamp'],
                             }

                new_entry['speed'] = fix_junos_speed(new_entry)
                entry_dict[new_entry['ifname']] = new_entry

                flags = lentry.get('if-config-flags',
                                   [{}])[0].get('iff-up', None)
                if flags is not None or (entry.get('type') == 'loopback'):
                    new_entry['state'] = 'up'
                else:
                    new_entry['state'] = 'down'

                if new_entry['mtu'] == 'Unlimited':
                    new_entry['mtu'] = 65536
                else:
                    new_entry['mtu'] = int(new_entry['mtu'])

                for elem in addrlist or []:
                    laddr = elem.get("ifa-local")[0]["data"]
                    if ':' in laddr:
                        plen = (elem.get("ifa-destination",
                                         [{"data": "0/128"}])[0]
                                ["data"].split("/")[1])
                        v6addresses.append(f'{laddr}/{plen}')
                    else:
                        plen = (elem.get("ifa-destination",
                                         [{"data": "0/32"}])[0]
                                ["data"].split("/")[1])
                        v4addresses.append(f'{laddr}/{plen}')
                vlanName = lentry.get('irb-domain', [{}])[0] \
                    .get('irb-bridge', [{}])[0] \
                    .get('data', '')
                if not vlanName:
                    new_entry['vlanName'] = vlanName
                else:
                    new_entry['vlanName'] = '-'
                new_entry['ip6AddressList'] = v6addresses
                new_entry['ipAddressList'] = v4addresses

                new_entries.append(new_entry)
            entry.pop('_logIf', [])

        if drop_indices:
            processed_data = np.delete(processed_data, drop_indices).tolist()

        if new_entries:
            processed_data.extend(new_entries)

        return processed_data

    def _clean_nxos_data(self, processed_data, raw_data):
        """Complex cleanup of NXOS interface data"""

        def fix_nxos_speed(entry):
            speed = str(self._common_speed_field_value(entry))
            if isinstance(speed, str):
                speed = speed.strip()
                if speed.endswith('(D)') or speed.endswith('(I)'):
                    speed = speed[:-3]
                if speed.startswith("unknown enum") or (speed == "auto"):
                    speed = str(self._get_missing_speed_value(entry))
                elif speed.startswith('a-'):
                    speed = speed[2:]

                if speed.endswith('G'):
                    speed = int(speed[:-1])*1000
                elif speed.endswith('Kbit'):
                    speed = int(speed.split()[0])/1000
                elif speed == '--':
                    speed = 0
                else:
                    speed = int(speed or 0)

            return speed

        new_entries = []
        unnum_intf = {}
        drop_indices = []
        entry_dict = {}

        unnum_intf_entry_idx = []  # backtrack to interface to fix

        for entry_idx, entry in enumerate(processed_data):
            if entry.get('_entryType', '') == 'mtumac':
                old_entry_idx = entry_dict.get(entry['ifname'], None)
                if old_entry_idx is not None:
                    old_entry = processed_data[old_entry_idx]
                    admin_state = entry['adminState'].lower()
                    if admin_state:
                        if admin_state == "administratively":
                            admin_state = "down"
                        old_entry['adminState'] = admin_state

                    old_entry['mtu'] = entry.get('mtu', 0) or old_entry['mtu']
                    macaddr = entry.get('macaddr', '')
                    if macaddr and not old_entry.get('_anycastMac', ''):
                        # VLAN MAC addr MUST not be replaced due to fabric SVI
                        old_entry['macaddr'] = convert_macaddr_format_to_colon(
                            macaddr)

                    old_entry['numChanges'] = entry['numChanges'] or 0
                    if entry.get('speed', ''):
                        old_entry['speed'] = fix_nxos_speed(entry)

                    lastChange = entry.get('statusChangeTimestamp', '')
                    if lastChange:
                        if any(x in lastChange for x in 'dwmy'):
                            lastChange = f'{lastChange} hours ago'

                        lastChange = parse(
                            lastChange,
                            settings={'RELATIVE_BASE': datetime.fromtimestamp(
                                (raw_data[0]['timestamp'])/1000),
                                'TIMEZONE': 'UTC'})
                    if lastChange:
                        old_entry['statusChangeTimestamp'] = int(
                            lastChange.timestamp() * 1000)
                    else:
                        old_entry['statusChangeTimestamp'] = 0
                    old_entry['description'] = entry.get('description', '')
                else:
                    self.logger.error(f'{raw_data[0]["hostname"]}: Unable to '
                                      f'find entry for {entry["ifname"]}')
                drop_indices.append(entry_idx)
                continue

            if entry.get('_entryType', '') == "vrf":
                # VRF is not an interface on NXOS, we're doing this to keep
                # it consistent with other NOS. So fill in dummy defaults.
                entry['type'] = 'vrf'
                entry['mtu'] = 1500
                entry['vni'] = 0

            ifname = expand_nxos_ifname(entry['ifname'])
            entry['ifname'] = ifname

            ipaddresses = entry.get('ipAddressList', None)
            if ipaddresses and ipaddresses != "--":
                pri_ipaddr = f"{entry['ipAddressList']}/{entry['_maskLen']}"
                ipaddr = [pri_ipaddr]
                for i, elem in enumerate(entry.get('_secIPs', [])):
                    if elem:
                        ipaddr.append(f"{elem}/{entry['_secmasklens'][i]}")
                entry['ipAddressList'] = ipaddr
            else:
                entry['ipAddressList'] = []

            if entry.get('ip6AddressList', None):
                if '_linklocal' in entry:
                    entry['ip6AddressList'].append(entry['_linklocal'])
            else:
                entry['ip6AddressList'] = []

            if ifname in entry_dict:
                # Older NXOS has this issue of providing loopback interfaces
                # with shortened names such as Lo0 while the IP addr section
                # lists them with full status such as loopback0. We have to
                # therefore manually merge them
                ifentry = processed_data[entry_dict[ifname]]
                if ifentry.get('type', '') == 'loopback':
                    ifentry['ipAddressList'] = entry['ipAddressList']
                    ifentry['master'] = entry['vrf']
                    ifentry['mtu'] = entry['mtu']
                if entry.get('_child_intf', []):
                    unnum_intf[ifname] = [pri_ipaddr]
                drop_indices.append(entry_idx)
                continue

            entry["statusChangeTimestamp1"] = entry.get(
                "statusChangeTimestamp", '')

            if entry.get('vrf', ''):
                entry['master'] = entry['vrf']

            if 'routeDistinguisher' in entry:
                # This is a VRF entry
                entry['macaddr'] = "00:00:00:00:00:00"
                entry['adminState'] = entry.get("state", "up").lower()
                entry['state'] = entry.get('state', 'down').lower()
                entry['speed'] = 0
                continue

            entry_dict[ifname] = entry_idx

            if 'reason' in entry:
                if entry['reason'] is not None:
                    entry['reason'] = entry['reason'].lower()

                if entry['reason'] == 'none' or not entry['reason']:
                    entry['reason'] = ''

                if entry['reason'] in ["link not connected",
                                       "xcvr not inserted"]:
                    entry['state'] = 'notConnected'

            if entry.get('reason', '') == 'administratively down':
                entry['adminState'] = 'down'
            else:
                entry['adminState'] = 'up'
            portmode = entry.get('_portmode', '')
            if portmode in ['access', 'trunk']:
                entry['master'] = 'bridge'

            portchan = entry.get('_portchannel', 0)
            if portchan and (isinstance(portchan, int) or
                             portchan.isnumeric()):
                entry['master'] = f'port-channel{portchan}'
                entry['type'] = 'bond_slave'

            if ifname.startswith('port-channel'):
                entry['type'] = 'bond'

            if not entry.get('macaddr', ''):
                entry['macaddr'] = "00:00:00:00:00:00"

            if entry.get('_anycastMac', ''):
                entry['interfaceMac'] = entry.get('macaddr', '')
                entry['macaddr'] = entry['_anycastMac']
                if entry.get('_forwardMode', '') != 'Anycast':
                    entry['reason'] += ', Fabric forwarding mode not enabled'

            entry['macaddr'] = convert_macaddr_format_to_colon(
                entry.get('macaddr', '0000.0000.0000'))

            unnum_if_parent = entry.get('_unnum_intf', '')
            if unnum_if_parent:
                if unnum_if_parent in unnum_intf:
                    # IPv6 has link local, so unnumbered is a v4 construct
                    entry['ipAddressList'] = unnum_intf[unnum_if_parent]
                else:
                    unnum_intf_entry_idx.append(entry_idx)

            if entry.get('_child_intf', []):
                unnum_intf[ifname] = [pri_ipaddr]

            if ifname == "mgmt0":
                entry['type'] = "ethernet"

            entry['type'] = entry.get('type', '').lower()
            if entry['type'] == 'eth':
                entry['type'] = 'ethernet'

            if 'ethernet' in entry.get('type', ''):
                entry['type'] = 'ethernet'

            if entry.get('vlan', '') in ["monitor", "--"]:
                entry['vlan'] = 0
            if ifname.startswith('Vlan'):
                entry['type'] = 'vlan'
            elif re.search(r'\.\d+$', ifname):
                entry['type'] = 'subinterface'
            elif ifname.startswith('nve'):
                entry['type'] = 'vxlan'
                entry['master'] = 'bridge'
            elif ifname.startswith('loopback'):
                entry['type'] = 'loopback'
            elif ifname.startswith('tunnel-te'):
                entry['type'] = 'tunnel-te'
            elif ifname.startswith('pw'):
                entry['type'] = 'pseudowire'

            if entry['type'] == 'vlan' and entry['ifname'].startswith('Vlan'):
                entry['vlan'] = int(entry['ifname'].split('Vlan')[1])

            entry['speed'] = fix_nxos_speed(entry)
            # have this at the end to avoid messing up processing

        # Fix unnumbered interface references
        for idx in unnum_intf_entry_idx:
            entry = processed_data[idx]
            entry['ipAddressList'] = unnum_intf.get(entry['_unnum_intf'], [])

        if drop_indices:
            processed_data = np.delete(processed_data, drop_indices).tolist()

        if new_entries:
            processed_data.extend(new_entries)

        return processed_data

    def _clean_linux_data(self, processed_data, _):
        """Pluck admin state from flags"""
        for entry in processed_data:
            if entry['type'] == 'ether':
                entry['type'] = 'ethernet'
            entry['state'] = entry['state'].lower()
            if entry['state'] == 'unknown':
                entry['state'] = 'up'  # loopback
            if ',UP' in entry['_flags']:
                entry['adminState'] = 'up'
            else:
                entry['adminState'] = 'down'

            entry['speed'] = self._textfsm_valid_speed_value(entry)

            # Linux interface output has no status change timestamp

        return processed_data

    def _clean_iosxr_data(self, processed_data, raw_data):

        entry_dict = {}
        devtype = raw_data[0].get('devtype', 'iosxr')
        drop_indices = []

        for i, entry in enumerate(processed_data):

            if entry.get('_entryType', '') == 'vrf':
                entry['master'] = ''
                entry['type'] = 'vrf'
                entry['mtu'] = -1
                entry['state'] = entry['adminState'] = 'up'
                entry['macaddr'] = "00:00:00:00:00:00"
                entry['speed'] = NO_SPEED
                continue

            if 'type' not in entry:
                # We have some weird outputs when some ephemeral interfaces
                # such as NDE_0 popup between the commands we run causing us
                # to not have full data about this interface. Drop such
                # interfaces. If they're not ephemeral, we'll get them in the
                # next run
                drop_indices.append(i)
                continue

            ifname = entry['ifname']
            if entry.get('_bondMbrs', ''):
                bond_mbrs = ' '.join(entry['_bondMbrs'])
            else:
                bond_mbrs = ''
            state = entry.get('state', '')
            if 'up' in state:
                # IOSVL2 images show up as up (connected)
                entry['state'] = 'up'
            elif 'down' in state:
                if 'notconnect' in state:
                    entry['state'] = 'notConnected'
                elif 'Autostate Enabled' in state:
                    entry['state'] = 'down'
                    entry['reason'] = state.split(',')[1].strip()
                elif 'disabled' in state:
                    entry['state'] = 'errDisabled'
                elif 'monitoring' in state:
                    entry['state'] = 'down'
                elif 'suspended' in state:
                    entry['state'] = 'down'
                    if not entry.get('reason', ''):
                        entry['reason'] = 'suspended'

            iftype = entry.get('type', 'ethernet').lower()
            if '.' in entry.get('ifname', ''):
                iftype = 'subinterface'
            elif bond_mbrs or iftype == "etherchannel":
                iftype = 'bond'
            elif iftype in ['ethernet', 'igbe', 'csr']:
                iftype = 'ethernet'
            elif iftype.endswith('gige'):
                iftype = 'ethernet'
            elif iftype.endswith('ge'):
                # Is this safe, assuming just ge ending means GigE?
                iftype = 'ethernet'
            elif iftype.endswith('ethernet'):
                iftype = 'ethernet'
            elif ifname.startswith('tunnel-te'):
                iftype = 'tunnel-te'
            elif ifname.startswith('pw'):
                iftype = 'pseudowire'
            elif ('10ge' in iftype or 'xge' in iftype or
                  'ethernet' in iftype or 'mb 802.3' in iftype or
                  'rp management' in iftype):
                iftype = 'ethernet'
            elif "embedded service engine" in iftype:
                iftype = 'svc-engine'
            elif "async group serial" in iftype:
                iftype = 'async-serial-grp'

            # More iftype processing below before setting it in entry
            speed = self._textfsm_valid_speed_value(entry)
            if speed != MISSING_SPEED:
                speed = int(speed)/1000  # is in Kbps
            entry['speed'] = speed

            if iftype == 'bond' and bond_mbrs:
                for mbr in bond_mbrs.split():
                    mbr = mbr.strip()
                    # expand the member name
                    mbr = expand_ios_ifname(mbr)
                    if mbr in entry_dict:
                        mbr_entry = entry_dict[mbr]
                        mbr_entry['type'] = 'bond_slave'
                        mbr_entry['master'] = entry['ifname']
                    else:
                        entry_dict[mbr] = {'master': entry['ifname'],
                                           'type': 'bond_slave'}

            if entry['adminState'] == 'administratively down':
                entry['state'] = 'down'
                entry['adminState'] = 'down'

            entry['macaddr'] = convert_macaddr_format_to_colon(
                entry.get('macaddr', '0000.0000.0000'))
            if not entry['macaddr']:
                entry['macaddr'] = '00:00:00:00:00:00'
            if entry['type'] == 'null':
                entry['macaddr'] = "00:00:00:00:00:00"
            entry['interfaceMac'] = convert_macaddr_format_to_colon(
                entry.get('interfaceMac', '0000.0000.0000'))

            lastChange = parse(
                entry.get('statusChangeTimestamp', ''),
                settings={'RELATIVE_BASE':
                          datetime.fromtimestamp(
                              (raw_data[0]['timestamp'])/1000), })
            if lastChange:
                entry['statusChangeTimestamp'] = int(lastChange.timestamp()
                                                     * 1000)
            if 'ipAddressList' not in entry:
                entry['ipAddressList'] = []
                entry['ip6AddressList'] = []
            elif ':' in entry['ipAddressList']:
                entry['ip6AddressList'] = entry['ipAddressList']
                entry['ipAddressList'] = []
            elif devtype == 'iosxr':
                entry['ip6AddressList'] = []

            if entry['ifname'].startswith('Vlan'):
                iftype = 'vlan'
                if entry.get('vlan', '') == '':
                    entry['vlan'] = entry['ifname'].split('Vlan')[1].strip()
            if entry.get('vlan', '') and entry.get('innerVlan', ''):
                iftype = "qinq"

            entry['type'] = iftype
            # This is specifically for IOSXE/IOS devices where
            # the IPv6 address uses capital letters
            if devtype != 'iosxr':
                entry['ip6AddressList'] = [x.lower()
                                           for x in entry.get('ip6AddressList',
                                                              [])]

            if entry['ifname'].endswith('.0'):
                entry['vlan'] = 0
            elif entry.get('vlan', '') == '':
                if entry['ipAddressList'] == entry['ip6AddressList'] == []:
                    entry['vlan'] = 1

            if entry['ipAddressList'] == 'Unknown':
                entry['ipAddressList'] = []
                entry['ip6AddressList'] = []

            if entry['ipAddressList'] or entry['ip6AddressList']:
                entry['master'] = entry.get('vrf', '')

            if entry['ifname'] in entry_dict:
                add_info = entry_dict[entry['ifname']]
                entry['master'] = add_info.get('master', '')
                entry['type'] = add_info.get('type', '')
            else:
                entry_dict[entry['ifname']] = entry

        if drop_indices:
            processed_data = np.delete(processed_data, drop_indices).tolist()

        return processed_data

    def _clean_iosxe_data(self, processed_data, raw_data):
        return self._clean_iosxr_data(processed_data, raw_data)

    def _clean_ios_data(self, processed_data, raw_data):
        return self._clean_iosxr_data(processed_data, raw_data)

    def _clean_sonic_data(self, processed_data, raw_data):
        return self._clean_cumulus_data(processed_data, raw_data)

    def _common_data_cleaner(self, processed_data, _):
        for entry in processed_data:
            entry['state'] = entry['state'].lower()

        return processed_data

    def _clean_panos_data(self, processed_data, _):
        mtu_data = {}
        interfaces = {}
        for entry in processed_data:
            _mtu_data = entry.get("_mtu_data")
            iftype = entry.get("type")
            master = entry.get("master")
            speed = entry.get("speed")
            state = entry.get("state")
            ipAddressList = entry.get("ipAddressList", [])
            ip6AddressList = entry.get("ip6AddressList", [])
            _ip = entry.get("_ip", "")

            # mtu values are collected separatly
            if _mtu_data:
                # fix json so that it can be parsed
                d = _mtu_data.split(": ", 1)[1].replace("'", "\"")
                d = re.sub(
                    r"([a-fA-F0-9]{2}(:[a-fA-F0-9]{2}){5})", r'"\1"', d)
                d = re.sub(r"(\"[\w0-9\.\/]+\": \{\s\},\s)", r"", d)
                d = re.sub(r"(,\s\})", r" }", d)
                j = loads(d)
                for ifname, value in j.items():
                    mtu_data[ifname] = value["mtu"]
                continue

            if entry["ifname"] in mtu_data:
                entry["mtu"] = mtu_data[entry["ifname"]]
            else:
                entry["mtu"] = ""

            # only hw if specify type value
            if iftype:
                if iftype == '0':
                    entry["type"] = "ethernet"
                elif iftype == '3':
                    entry["type"] = "vlan"
                elif iftype == '5':
                    entry["type"] = "loopback"
                elif iftype == '6':
                    entry["type"] = "tunnel"
                else:
                    # not sure if it's better to put unknown
                    entry["type"] = ""

            if master:
                if master != "N/A":
                    entry["master"] = master.split(":")[1]
                else:
                    entry["master"] = "default"

            # speed is not always provided
            if speed and speed == "[n/a]":
                # if it is a physical interface, it's not good
                if entry["type"] == "ethernet":
                    entry["speed"] = MISSING_SPEED
                else:
                    entry["speed"] = NO_SPEED

            # this happens when there's only one entry
            if isinstance(ipAddressList, str):
                entry["ipAddressList"] = [ipAddressList]

            if isinstance(ip6AddressList, str):
                entry["ip6AddressList"] = [ip6AddressList]

            if _ip and _ip != "N/A" and _ip != "unknown":
                if ":" in _ip:
                    entry["ip6AddressList"] = entry.get("ip6AddressList", []) \
                        + [_ip]
                else:
                    entry["ipAddressList"] = entry.get("ipAddressList", []) \
                        + [_ip]
            if state and "/" in state:
                # clear from [n/a] before splitting
                entry['state'] = state.replace("[n/a]", "").split("/")[2]

            if "management" in entry["ifname"].lower():
                entry["state"] = "up"
                entry["type"] = "ethernet"

            if entry.get("type") == "ethernet":
                # physical interface, save data to use with subinterfaces
                interfaces[entry["ifname"]] = {}
                interfaces[entry["ifname"]]["mtu"] = entry["mtu"]
                interfaces[entry["ifname"]]["speed"] = entry["speed"]
                interfaces[entry["ifname"]]["state"] = entry["state"]
                interfaces[entry["ifname"]]["adminState"] = entry["adminState"]

            if "." in entry["ifname"]:
                parentIf, _ = entry["ifname"].split(".")
                if parentIf in interfaces:
                    entry["mtu"] = interfaces[parentIf]["mtu"]
                    entry["state"] = interfaces[parentIf]["state"]
                    entry["adminState"] = interfaces[parentIf]["adminState"]
                    entry["speed"] = interfaces[parentIf]["speed"]
                    entry["type"] = "subinterface"

        # remove mtu data cmd
        if len(processed_data) > 1:
            processed_data.pop(0)
        return processed_data



"""
This module contains the logic of the plugin in charge of importing an
inventory from an Ansible inventory file.
"""
# pylint: disable=no-self-argument

from pathlib import Path
from typing import Dict, Union
import logging
from pydantic import Field, validator
import yaml
from suzieq.poller.controller.source.base_source import Source, SourceModel

logger = logging.getLogger(__name__)


class AnsibleSourceModel(SourceModel):
    """Model used to validate ansible input format
    """
    hosts: Union[str, Dict] = Field(alias='path')

    @validator('hosts')
    def validate_and_set(cls, path: str):
        """checks if the path is valid
        """
        if isinstance(path, str):
            if not Path(path).is_file():
                raise ValueError(
                    f"No file found at {path}")

            try:
                with open(path, 'r') as f:
                    inventory = yaml.safe_load(f)
            except Exception as error:
                raise ValueError(
                    f'Unable to process Ansible inventory: {str(error)}'
                )

        elif isinstance(path, Dict):
            inventory = path
        if '_meta' not in inventory \
                or 'hostvars' not in inventory['_meta']:
            if isinstance(inventory, list) and 'namespace' in inventory[0]:
                raise ValueError(
                    'Invalid Ansible inventory, found Suzieq inventory'
                )
            raise ValueError(
                'Invalid Ansible inventory, missing keys: _meta and / or'
                "hostvars\n \tUse 'ansible-inventory --list' to create "
                'the correct file'
            )

        return inventory['_meta']['hostvars']


class AnsibleInventory(Source):
    """The AnsibleInventory is a class allowing to import the output
    of 'ansible-inventory command' as input of the poller.
    """

    @classmethod
    def get_data_model(cls):
        return AnsibleSourceModel

    def _load(self, input_data):
        # pydantic aliases mess up things
        # put back data in 'hosts' from the 'path' alias
        if not self._validate:
            input_data['hosts'] = input_data.pop('path', {})

        super()._load(input_data)
        inventory = self._get_inventory()
        self.set_inventory(inventory)

    def _get_inventory(self) -> Dict:
        """Parse the output of ansible-inventory command for processing.

        Ansible pulls together the inventory information from multiple files.
        The information relevant to sq-poller maybe present in different files
        as different vars. ansible-inventory command luckily handles this for
        us. This function takes the JSON output of that command and gathers
        the data needed to start polling.

        Returns:
            Dict: A list containing a dictionary of data with the data to
                connect to the host.
        """

        in_hosts = self._data.hosts
        out_inv = {}
        for host in in_hosts:
            entry = in_hosts[host]

            ansible_host = entry.get('ansible_host')
            if not ansible_host:
                logger.warning(f'{self.name} skipping ansible device '
                               'without hostname')
                continue

            ansible_user = entry.get('ansible_user')
            if not ansible_user:
                logger.warning(
                    f'{self.name} skipping ansible device without username')
                continue

            # Get password if any
            password = None
            if 'ansible_password' in entry:
                password = entry['ansible_password']

            # Retrieve password information
            devtype = None
            if entry.get('ansible_network_os') in ['eos', 'panos']:
                devtype = entry.get('ansible_network_os')
                transport = 'https'
                port = 443
            else:
                transport = 'ssh'
                port = entry.get('ansible_port', 22)

            # Get keyfile
            keyfile = entry.get('ansible_ssh_private_key_file', None)
            if keyfile and not Path(keyfile).exists():
                logger.warning(
                    f"{self.name} Ignored host {ansible_host} because "
                    f"associated keyfile {keyfile} does not exist"
                )
                continue

            host = {
                'address': ansible_host,
                'username': ansible_user,
                'port': port,
                'password': password,
                'transport': transport,
                'devtype': devtype,
                'namespace': self._namespace,
                'ssh_keyfile': keyfile,
                'hostname': None
            }
            out_inv[f"{self._namespace}.{ansible_host}.{port}"] = host

        return out_inv



import asyncio
import copy
import json
import logging
import operator
import os
import time
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timezone
from http import HTTPStatus
from tempfile import mkstemp
from typing import Dict, List

import pyarrow as pa
import yaml
from packaging import version as version_parse

from suzieq.poller.worker.services.svcparser \
    import cons_recs_from_json_template
from suzieq.shared.sq_plugin import SqPlugin
from suzieq.shared.utils import known_devtypes
from suzieq.version import SUZIEQ_VERSION

# How long b4 declaring node dead
HOLD_TIME_IN_MSECS = 60000
# How many unsuccessful polls before marking records with active=false
HYSTERESIS_FAILURE_CNT = 3


@dataclass
class RsltToken:
    '''Cookie passed between service and node objects'''
    start_time: int    # When this cmd was first posted to node
    nodename: str      # Name of node, used to get poller cb again
    bootupTimestamp: int
    nodeQsize: int        # Size of the node q at queuing time
    service: str          # Name of this service, if node needs it
    timeout: int          # timeout value for cmd to complete


@dataclass
class ServiceStats:
    '''Capture useful stats about service'''
    total_time: List[float] = field(default_factory=list)
    gather_time: List[float] = field(default_factory=list)
    svcQsize: List[float] = field(default_factory=list)
    nodeQsize: List[float] = field(default_factory=list)
    wrQsize: List[float] = field(default_factory=list)
    rxBytes: List[float] = field(default_factory=list)
    empty_count: int = 0
    time_excd_count: int = 0    # Number of times total_time > poll period
    next_update_time: int = 0   # When results will be logged


class Service(SqPlugin):
    '''Main class for handling various services/tables processing on devices'''

    def get_poller_schema(self):
        '''Return the schema used by this service'''
        return self._poller_schema

    def set_poller_schema(self, value: dict):
        '''Set the poller schema used by this service'''
        self._poller_schema = value

    def __init__(self, name, defn, period, stype, keys, ignore_fields, schema,
                 queue, db_access, run_once=None):
        self.name = name
        self.defn = defn
        self.ignore_fields = ignore_fields or []
        self.writer_queue = queue
        self.keys = keys
        self.schema = schema.get_arrow_schema()
        self.schema_table = schema
        self.period = period
        self.stype = stype
        self.logger = logging.getLogger(__name__)
        self.run_once = run_once
        self.post_timeout = 5
        self.sigend = False
        self.version = schema.version
        # Get sqobject to retrieve the data of this service
        self._db_access = db_access

        self.update_nodes = False  # we have a new node list
        self.rebuild_nodelist = False  # used only when a node gets init
        self.node_postcall_list = {}
        self.new_node_postcall_list = {}
        self.previous_results = {}
        self._poller_schema = {}
        self.node_boot_times = defaultdict(int)
        self._failed_node_set = set()
        self._consecutive_failures = defaultdict(int)

        self.poller_schema = property(
            self.get_poller_schema, self.set_poller_schema)
        self.poller_schema_version = None

        # The queue to which all nodes will post the result of the command
        self.result_queue = asyncio.Queue()

        # Add the hidden fields to ignore_fields
        self.ignore_fields.append("timestamp")

        if "namespace" not in self.keys:
            self.keys.insert(0, "namespace")

        if "hostname" not in self.keys:
            self.keys.insert(1, "hostname")

        self.partition_cols = schema.get_partition_columns()

        # Setup dictionary of NOS specific extracted data cleaners
        self.dev_clean_fn = {}
        common_dev_clean_fn = getattr(self, '_common_data_cleaner', None)
        for x in known_devtypes():
            if x.startswith('junos'):
                dev = 'junos'
            else:
                dev = x
            self.dev_clean_fn[x] = getattr(
                self, f'_clean_{dev}_data', None) or common_dev_clean_fn

    @ staticmethod
    def is_status_ok(status: int) -> bool:
        '''Did the node return a successful command output'''
        if status in [0, HTTPStatus.OK]:
            return True
        return False

    def get_data(self):
        """provide the data that is interesting for a service

        does not include queue or schema"""

        r = {}
        for fld in ('name defn ignore_fields keys period stype '
                    'partition_cols').split(' '):
            r[field] = getattr(self, fld)
        # textFSM objects can't be jsoned, so changing it
        for device in r['defn']:
            if 'textfsm' in r['defn'][device]:
                r['defn'][device]['textfsm'] = str(
                    r['defn'][device]['textfsm'])
        return r

    async def set_nodes(self, node_call_list):
        """New node list for this service"""
        if self.node_postcall_list:
            self.new_node_postcall_list = node_call_list
            self.update_nodes = True
        else:
            self.node_postcall_list = node_call_list

    def _get_default_vals(self) -> dict:
        '''Get default values based on type'''
        return({
            pa.string(): "",
            pa.int32(): 0,
            pa.int64(): 0,
            pa.float32(): 0.0,
            pa.float64(): 0.0,
            pa.date64(): 0.0,
            pa.bool_(): False,
            pa.list_(pa.string()): [],
            pa.list_(pa.int64()): [],
        })

    def get_empty_record(self):
        '''Return an empty record matching schema'''
        map_defaults = self._get_default_vals()

        defaults = [map_defaults[x] for x in self.schema.types]
        rec = dict(zip(self.schema.names, defaults))

        return rec

    def _dump_output(self, outputs) -> str:
        """Dump the output to a YAML file"""
        if outputs:
            yml = {"service": self.name,
                   "type": self.run_once,
                   "output": outputs}
            fd, name = mkstemp(suffix=f"-{self.name}-poller.yml")
            f = os.fdopen(fd, "w")
            f.write(yaml.dump(yml))
            f.close()
            return name
        return ""

    def get_diff(self, old, new):
        """Compare list of dictionaries ignoring certain fields
        Return list of adds and deletes
        """
        # If old is empty there is no need to look for differences
        # just return adds
        if not old:
            return new, []

        # Work out the differences between the new and the old records
        adds = []
        dels = []
        koldvals = []
        knewvals = []
        koldkeys = []
        knewkeys = []

        fieldsToCheck = set()
        # keys that start with _ are transient and must be ignored
        # from comparison
        for elem in new:
            # Checking the __iter__ attribute to check if it is iterable
            # is not pythonic but much faster, we need to be fast since this
            # piece of code is executed tons of times
            vals = {k: set(v) if hasattr(v, '__iter__') and
                    not isinstance(v, str) and
                    not isinstance(v, dict) else str(v)
                    for k, v in elem.items()
                    if k not in self.ignore_fields and not k.startswith('_')
                    and k in self.schema_table.fields}
            kvals = {v for k, v in elem.items() if k in self.keys}
            knewvals.append(vals)
            knewkeys.append(kvals)
            fieldsToCheck.update(vals.keys())

        for elem in old:
            # Get from the old records only the fields there are in new records
            # since we want to compare what we have, also excluding all the
            # derived derived colums, which are not explicitly written in
            # the data.
            vals = {k: set(v) if hasattr(v, '__iter__') and
                    not isinstance(v, str) and
                    not isinstance(v, dict) else str(v)
                    for k, v in elem.items()
                    if k in fieldsToCheck}
            kvals = {v for k, v in elem.items() if k in self.keys}
            koldvals.append(vals)
            koldkeys.append(kvals)

        adds = [new[k] for k, v in enumerate(knewvals) if v not in koldvals]
        dels = [old[k] for k, v in enumerate(koldvals)
                if v not in knewvals and koldkeys[k] not in knewkeys]

        # The sqvers type for dels might be float, we expect str
        for d in dels:
            if d.get('sqvers'):
                d['sqvers'] = str(d['sqvers'])

        if adds and self.stype == "counters":
            # If there's a change in any field of the counters, update them all
            # simplifies querying
            adds = new

        return adds, dels

    def textfsm_data(self, raw_input, fsm_template, entry_type, _1, _2):
        """Convert unstructured output to structured output"""

        records = []
        fsm_template.Reset()
        res = fsm_template.ParseText(raw_input)

        for entry in res:
            metent = dict(zip(fsm_template.header, entry))
            if entry_type:
                metent['_entryType'] = entry_type
            records.append(metent)

        return records

    async def post_results(self, result, token) -> None:
        """Callback fn used to post the results back to the service"""
        if self.result_queue:
            self.result_queue.put_nowait((token, result))
        else:
            self.logger.error(f"No queue for service {self.name}")

    def call_node_postcmd(self, postcall, nodename) -> None:
        """Start data gathering by calling the post command list"""
        if postcall and postcall['postq']:
            token = RsltToken(int(time.time()*1000), nodename, 0, 0, self.name,
                              self.period-5)
            postcall['postq'](self.post_results, self.defn, token)

    def clean_json_input(self, data):
        """Clean the JSON input data that is sometimes messed up
        Each service can implement its own version of the cleanup
        """
        return data.get('data', {})

    async def start_data_gather(self) -> None:
        """Start data gathering by calling the post command list
        This is only used to fire up the calls the first time. After this,
        each node gets into a self-adapting curve of calling at least after
        the duration specified for the service period. If nodes are slower,
        they'll be scheduled to run after the specified service period after
        their return. We could use this to track slow nodes.
        """

        for node in self.node_postcall_list:
            try:
                self.call_node_postcmd(self.node_postcall_list[node],
                                       node),
            except TimeoutError:
                self.logger.warning(f"Node post failed for {node}")
                continue

    def _get_def_for_version(self, defn: List, version: str) -> Dict:
        '''Extract the right definition for the version'''
        if isinstance(defn, dict):
            return defn

        if not isinstance(defn, list):
            return defn

        for item in defn:
            os_version = item.get('version', '')
            if os_version == "all":
                return item
            if version == 0:
                # There are various outputs that due to an old parsing bug
                # return a node version of 0. Use 'all' for those
                continue
            opdict = {'>': operator.gt, '<': operator.lt,
                      '>=': operator.ge, '<=': operator.le,
                      '=': operator.eq, '!=': operator.ne}
            op = operator.eq

            for elem, val in opdict.items():
                if os_version.startswith(elem):
                    os_version = os_version.replace(
                        elem, '').strip()
                    op = val
                    break

            if op(version_parse.LegacyVersion(version),
                  version_parse.LegacyVersion(os_version)):
                return item

        return None

    def _get_device_normalizer(self, data: Dict) -> Dict:
        '''Returns the normalizer string appropriate for the data

        Every NOS/version has a normalizer string specified in the
        service config file (under suzieq/config). This function
        returns that string for the given data input
        '''
        nfn = self.defn.get(data.get("hostname"), {})
        if not nfn:
            nfn = self.defn.get(data.get("devtype"), {})
        if nfn:
            # If we're riding on the coattails of another device
            # get that device's normalization function
            if isinstance(nfn, dict):
                copynfn = nfn.get("copy", {})
                if copynfn:
                    nfn = self.defn.get(copynfn, {})
            if isinstance(nfn, list):
                nfn = self._get_def_for_version(nfn, data.get("version"))

        return nfn

    def _process_each_output(self, elem_num, data, nfn):
        """Workhorse processing routine for each element in output"""

        result = []
        if (Service.is_status_ok(data.get('status', -1))):
            if not data["data"]:
                return result

            # Check host-specific normalization if any
            if nfn:
                norm_str = nfn.get("normalize", None)
                if norm_str is None and isinstance(nfn.get('command', None),
                                                   list):
                    # Can happen because we've a list of commands associated
                    # Pick the normalization function associated with this
                    # output
                    norm_str = nfn.get('command', [])[elem_num].get(
                        'normalize', None)
                if norm_str:
                    if isinstance(data["data"], str):
                        try:
                            in_info = json.loads(data["data"])
                        except json.JSONDecodeError:
                            in_info = self.clean_json_input(data)
                            if not in_info:
                                self.logger.error(
                                    "Received non-JSON output where "
                                    "JSON was expected for %s on "
                                    "node %s, %s",
                                    data["cmd"], data["hostname"],
                                    data["data"]
                                )
                                return result
                            try:
                                in_info = json.loads(in_info)
                            except json.JSONDecodeError:
                                self.logger.error(
                                    "Received non-JSON output where "
                                    "JSON was expected for %s on "
                                    "node %s, %s",
                                    data["cmd"], data["hostname"],
                                    data["data"]
                                )
                                return result
                    else:
                        in_info = data["data"]

                    if in_info:
                        # EOS' HTTP returns errors like this.
                        if isinstance(in_info, dict):
                            tmp = in_info.get('data', [])
                            if tmp and tmp[0].get('errors', {}):
                                return []

                        result = cons_recs_from_json_template(
                            norm_str, in_info)

                else:
                    tfsm_template = nfn.get("textfsm", None)
                    entry_type = None
                    if tfsm_template is None and isinstance(
                            nfn.get('command', None), list):
                        # Can happen because we've a list of cmds associated
                        # Pick the normalization function associated with this
                        # output
                        tfsm_template = nfn.get('command', [])[elem_num].get(
                            'textfsm', None)
                        entry_type = nfn.get('command', [])[elem_num].get(
                            '_entryType', None)
                    else:
                        entry_type = nfn.get('_entryType', None)

                    if not tfsm_template:
                        return result

                    in_info = []
                    if isinstance(data['data'], dict):
                        if "output" in data["data"]:
                            in_info = data["data"]["output"]
                        elif "messages" in data["data"]:
                            # This is Arista's bash output format
                            in_info = data["data"]["messages"][0]
                    else:
                        in_info = data['data']
                    if not in_info:
                        return result  # No data to process
                    # Clean data is invoked inside this due to the way we
                    # munge the data and force the types to adhere to the
                    # specified type
                    result = self.textfsm_data(
                        in_info, tfsm_template, entry_type, self.schema, data
                    )
            else:
                self.logger.error(
                    "%s: No normalization/textfsm function for device %s",
                    self.name, data["hostname"])

        return result

    def process_data(self, data):
        """Derive the data to be stored from the raw input"""
        result_list = []
        do_merge = True
        for i, item in enumerate(data):
            nfn = self._get_device_normalizer(item)
            do_merge = do_merge and nfn and nfn.get('merge', True)
            tmpres = self._process_each_output(i, item, nfn)
            result_list.append(tmpres)

        if do_merge:
            result = self.merge_results(result_list, data)
        else:
            result = [ele for sublist in result_list for ele in sublist]
        return self.clean_data(result, data)

    def get_key_flds(self):
        """Get the key fields associated with this service.
        Its a function because we want to override it.
        """
        return list(filter(lambda x: x not in ['namespace', 'hostname'],
                           self.keys))

    def merge_results(self, result_list, _):
        '''Merge the results from multiple commands into a single result'''

        int_res = {}
        keyflds = self.get_key_flds()

        for result in result_list:
            for entry in result:
                keyvals = []
                for kfld in keyflds:
                    if not entry.get(kfld, None):
                        keyvals.append('_default')
                    else:
                        val = entry.get(kfld, '')
                        if not isinstance(val, str):
                            val = str(val)
                        keyvals.append(val)

                if entry.get('_entryType', ''):
                    keyvals.append(entry['_entryType'])

                key = '-'.join(keyvals)

                if key not in int_res:
                    int_res[key] = entry
                else:
                    existing = int_res[key]
                    for fld in entry:
                        if entry[fld] and not existing.get(fld, None):
                            existing[fld] = entry[fld]
                        elif entry[fld]:
                            # cleanup routine specific to service can resolve
                            existing[fld+"-_2nd"] = entry[fld]

        final_res = list(int_res.values())

        return(final_res)

    def _get_devtype_from_input(self, input_data):
        if isinstance(input_data, list):
            return input_data[0].get('devtype', None)
        else:
            return input_data.get('devtype', None)

    def clean_data(self, processed_data, raw_data):
        """Massage the extracted data to match record specified by schema"""

        dev_clean_fn = self.dev_clean_fn.get(self._get_devtype_from_input(
            raw_data), None)
        if dev_clean_fn:
            processed_data = dev_clean_fn(processed_data, raw_data)

        return self.clean_data_common(processed_data, raw_data)

    def clean_data_common(self, processed_data, raw_data):
        """Fix the type and default value of of each extracted field

        This routine is common to all services. It ensures that all the missing
        fields, as defined by the schema, are added to the records extracted.
        Furthermore, each field is set to the specified type.
        """

        # Build default data structure
        schema_rec = {}
        def_vals = self._get_default_vals()

        ptype_map = {
            pa.string(): str,
            pa.int32(): int,
            pa.int64(): int,
            pa.float32(): float,
            pa.float64(): float,
            pa.date64(): float,
            pa.list_(pa.string()): list,
            pa.list_(pa.int64()): list,
            pa.bool_(): bool,
        }

        for fld in self.schema:
            default = def_vals[fld.type]
            schema_rec.update({fld.name: default})

        if isinstance(raw_data, list):
            read_from = raw_data[0]
        else:
            read_from = raw_data

        # pylint: disable=too-many-nested-blocks
        for entry in processed_data or []:
            entry.update({"hostname": read_from["hostname"]})
            entry.update({"namespace": read_from["namespace"]})
            entry.update({"timestamp": read_from["timestamp"]})
            entry.update({"sqvers": self.version})
            for fld, val in schema_rec.items():
                if fld not in entry:
                    if fld == "active":
                        entry.update({fld: True})
                    else:
                        entry.update({fld: val})
                else:
                    fld_type = self.schema.field(fld).type
                    if not isinstance(entry[fld], ptype_map[fld_type]):
                        try:
                            entry[fld] = ptype_map[fld_type](entry[fld])
                        except (ValueError, TypeError):
                            entry[fld] = val
                    elif isinstance(entry[fld], list):
                        for i, ele in enumerate(entry[fld]):
                            if not isinstance(ele,
                                              ptype_map[fld_type.value_type]):
                                try:
                                    if ptype_map[fld_type.value_type] == int:
                                        entry[fld][i] = int(entry[fld][i])
                                    elif ptype_map[fld_type.value_type] == str:
                                        entry[fld][i] = str(entry[fld][i])
                                    else:
                                        raise ValueError
                                except (ValueError, TypeError):
                                    entry[fld][i] = val
        return processed_data

    async def commit_data(self, result: Dict, namespace: str, hostname: str):
        """Write the result data out"""
        records = []
        key = f'{namespace}.{hostname}'
        prev_res = self.previous_results.get(key, None)
        # No data previously polled with this service from the the
        # current namespace and hostname. Check if the datastore contains some
        # information about
        filter_df = f"hostname.str.match('{hostname}') and " \
                    f"namespace.str.match('{namespace}')"
        if prev_res is None:
            df = self._db_access.read(
                self.schema_table._table,
                'pandas',
                start_time='',
                end_time='',
                columns=self.schema_table.fields,
                view='latest',
                key_fields=self.schema_table.key_fields(),
                add_filter=filter_df,
                hostname=[hostname],
                namespace=[namespace]).query('active')
            prev_res = df.to_dict('records')

        if result or prev_res:
            adds, dels = self.get_diff(prev_res, result)
            if adds or dels:
                self.previous_results[key] = copy.deepcopy(result)
                for entry in adds:
                    records.append(entry)
                for entry in dels:
                    if entry.get("active", True):
                        # If there's already an entry marked as deleted
                        # No point in adding one more
                        entry.update({"active": False})
                        entry.update(
                            {"timestamp":
                             int(datetime.now(tz=timezone.utc).timestamp()
                                 * 1000)}
                        )
                        records.append(entry)

                self._post_work_to_writer(records)

    def _post_work_to_writer(self, records: dict):
        """This posts the data to be written to the worker queue"""
        if records:
            self.writer_queue.put_nowait(
                {
                    "records": records,
                    "topic": self.name,
                    "schema": self.schema,
                    "partition_cols": self.partition_cols
                }
            )

    def reset_stats(self, stats: ServiceStats) -> None:
        """Reset the stats arrays.
        """
        stats.total_time = []
        stats.gather_time = []
        stats.svcQsize = []
        stats.nodeQsize = []
        stats.wrQsize = []
        stats.rxBytes = []
        stats.empty_count = 0
        stats.time_excd_count = 0

    def compute_basic_stats(self, statsList, newval: int) -> None:
        """Compute min/max/avg for given stats with the new value

        :param statsList: list consisting of 3 values in order: min, max, avg
        :type statsList: list_

        :param newval: The newval to update the list with
        :type newval: int
        """
        if not statsList:
            statsList = [newval, newval, newval]
            return statsList

        if newval < statsList[0]:
            statsList[0] = newval
        if newval > statsList[1]:
            statsList[1] = newval
        statsList[2] = (statsList[2]+newval)/2
        if statsList[2] < 0.001:
            statsList[2] = 0

        return statsList

    def update_stats(self, stats: ServiceStats, total_time: int,
                     gather_time: int, qsize: int, wrQsize: int,
                     nodeQsize: int, rxBytes) -> bool:
        """Update per-node stats"""
        write_stat = False
        now = int(time.time()*1000)

        stats.total_time = self.compute_basic_stats(stats.total_time,
                                                    total_time)
        stats.gather_time = self.compute_basic_stats(stats.gather_time,
                                                     gather_time)
        stats.svcQsize = self.compute_basic_stats(stats.svcQsize, qsize)
        stats.wrQsize = self.compute_basic_stats(stats.wrQsize, wrQsize)
        stats.nodeQsize = self.compute_basic_stats(stats.nodeQsize, nodeQsize)
        stats.rxBytes = self.compute_basic_stats(stats.rxBytes, rxBytes)

        if total_time > self.period*1000:
            stats.time_excd_count += 1
            write_stat = True

        if now > stats.next_update_time:
            stats.next_update_time = now + 5*60*1000  # 5 mins
            write_stat = True

        return write_stat

    def is_first_run(self, key: str) -> bool:
        '''Boolean indicating wether the current run
           is the first since the service is running.
        '''
        return not self.previous_results.get(key, None)

    # pylint: disable=too-many-statements
    async def run(self):
        """Start the service"""
        self.logger.info(f"running service {self.name} ")

        if self.run_once:
            total_nodes = len(self.node_postcall_list)
        else:
            total_nodes = 0
        # Fire up the initial posts
        await self.start_data_gather()
        loop = asyncio.get_running_loop()
        # pylint: disable=unnecessary-lambda
        pernode_stats = defaultdict(lambda: ServiceStats())

        while True:
            try:
                token, output = await self.result_queue.get()
            except asyncio.CancelledError:
                self.logger.warning(
                    f'Service: {self.name}: Received signal to terminate')
                return
            qsize = self.result_queue.qsize()

            self.logger.debug(f"Extracted response for {self.name} service")
            if isinstance(token, list):
                token = token[0]
            gather_time = int(time.time()*1000) - token.start_time

            status = HTTPStatus.NO_CONTENT        # Empty content
            write_poller_stat = False
            rxBytes = 0

            if output:
                ostatus = [x.get('status', -1) for x in output]

                should_commit = True

                # we should always commit during the first run of the service
                # for each node regardless their status.
                if not self.is_first_run(token.nodename):
                    # If it's not the first run and some command did fail,
                    # increment the count to be used in hysteresis.
                    # The info will be committed in a future run.
                    if any((not Service.is_status_ok(x) for x in ostatus)):
                        self._consecutive_failures[token.nodename] += 1
                    # otherwise reset the counter.
                    else:
                        self._consecutive_failures[token.nodename] = 0

                if 0 < self._consecutive_failures[token.nodename] < \
                        HYSTERESIS_FAILURE_CNT:
                    self.logger.info(
                        f"{self.name} {token.nodename} node failure hysteresis"
                        ", skipping data commit")
                    should_commit = False

                # pylint: disable=use-a-generator
                write_poller_stat = not all([Service.is_status_ok(x)
                                             for x in ostatus])
                status = ostatus[0]
                if (status in [0, 200]):
                    rxBytes = len(str(output))

                # We don't expect the output from two different hostnames
                nodename = output[0]["hostname"]
                self.logger.debug(f"Extracted response from {nodename} for "
                                  f"{self.name} service")
                # We can terminate processing if we've no data returned
                # and we're reading inputs from a file
                if nodename == '_filedata' and status == HTTPStatus.NO_CONTENT:
                    return

                # Process data independent of status to return an empty list
                # so that the output can be updated. For example, if a service
                #  is disabled, this can be the condition.
                if self.run_once == "gather":
                    if self.is_status_ok(status):
                        self._post_work_to_writer(json.dumps(output, indent=4))
                    total_nodes -= 1
                    if total_nodes <= 0:
                        self.logger.warning(
                            f'Service: {self.name}: Finished gathering data')
                        return
                    continue

                try:
                    result = self.process_data(output)
                except Exception:  # pylint: disable=broad-except
                    result = []
                    status = HTTPStatus.BAD_GATEWAY
                    write_poller_stat = True
                    self.logger.exception(
                        f'Processing data failed for service '
                        f'{self.name} on node {nodename}')

                # Don't write the error every time the failure happens
                if write_poller_stat:
                    if nodename in self._failed_node_set:
                        write_poller_stat = False
                    else:
                        self._failed_node_set.add(nodename)
                elif nodename in self._failed_node_set:
                    # So there was no error in this command that had failed b4
                    self._failed_node_set.remove(nodename)

                # If a node from init state to good state, hostname will change
                # So fix that in the node list
                if self.run_once == "process":
                    if self.is_status_ok(status):
                        self._post_work_to_writer(json.dumps(result, indent=4))
                    total_nodes -= 1
                    if total_nodes <= 0:
                        return
                    continue

                if should_commit:
                    await self.commit_data(result, output[0]["namespace"],
                                           output[0]["hostname"])
            elif self.run_once in ["gather", "process", "update"]:
                total_nodes -= 1
                if total_nodes <= 0:
                    self.logger.info(f'Service: {self.name}: Finished '
                                     f'{self.run_once}ing data')
                    return
                continue

            total_time = int(time.time()*1000) - token.start_time

            if output:

                stats = pernode_stats[token.nodename]
                write_poller_stat = (self.update_stats(
                    stats, total_time, gather_time, qsize,
                    self.writer_queue.qsize(), token.nodeQsize, rxBytes) or
                    write_poller_stat or
                    self.is_first_run(token.nodename))
                pernode_stats[token.nodename] = stats
                if write_poller_stat:
                    poller_stat = [
                        {"hostname": (output[0]["hostname"] or
                                      output[0]['address']),
                         "sqvers": self.poller_schema_version,
                         "namespace": output[0]["namespace"],
                         "active": True,
                         "service": self.name,
                         "status": status,
                         "svcQsize": stats.svcQsize,
                         "wrQsize": stats.wrQsize,
                         "nodeQsize": stats.nodeQsize,
                         "rxBytes": stats.rxBytes,
                         "pollExcdPeriodCount": stats.time_excd_count,
                         "gatherTime": stats.gather_time,
                         "totalTime": stats.total_time,
                         "version": SUZIEQ_VERSION,
                         "nodesPolledCnt": len(self.node_postcall_list),
                         "nodesFailedCnt": len(self._failed_node_set),
                         "timestamp": int(datetime.now(tz=timezone.utc)
                                          .timestamp() * 1000)}]

                    self.writer_queue.put_nowait(
                        {
                            "records": poller_stat,
                            "topic": "sqPoller",
                            "schema": self.poller_schema,
                            "partition_cols": self.partition_cols
                        })
                    self.reset_stats(pernode_stats[token.nodename])

            # Post a cmd to fire up the next poll after the specified period
            if self.run_once == "update":
                total_nodes -= 1
                if total_nodes <= 0:
                    self.logger.info(
                        f'Service: {self.name}: Finished updating data')
                    return
                continue
            self.logger.debug(
                f"Rescheduling service for {self.name} service")
            loop.call_later(self.period, self.call_node_postcmd,
                            self.node_postcall_list.get(token.nodename),
                            token.nodename)



from suzieq.sqobjects.basicobj import SqObject


class TopcpuObj(SqObject):
    '''The object providing access to the topcpu table'''

    def __init__(self, **kwargs):
        super().__init__(table='topcpu', **kwargs)
        self._valid_get_args = ['namespace',
                                'hostname', 'columns', 'query_str']
